# curio/queue.py
#
# A few different queue structures.

# -- Standard library

from collections import deque
from heapq import heappush, heappop
from concurrent.futures import Future
import threading
import socket as std_socket
import asyncio

# -- Curio

from .traps import _future_wait
from .sched import SchedFIFO, SchedBarrier
from .errors import CurioError, CancelledError
from .meta import awaitable, asyncioable
from . import workers

__all__ = ['Queue', 'PriorityQueue', 'LifoQueue', 'UniversalQueue']

class QueueBase:
    '''
    Base class for queues used to communicate between Curio tasks.
    Not safe for use with external threads or processes.
    '''

    def __init__(self, maxsize=0):
        self.maxsize = maxsize
        self._get_waiting = SchedFIFO()
        self._put_waiting = SchedFIFO()
        self._join_waiting = SchedBarrier()
        self._task_count = 0
        self._queue = self._init_internal_queue()

    def __repr__(self):
        res = super().__repr__()
        return '<%s, len=%d>' % (res[1:-1], len(self._queue))

    def empty(self):
        return not self._queue

    def full(self):
        return self.maxsize and len(self._queue) == self.maxsize

    async def get(self):
        must_wait = bool(self._get_waiting)
        while must_wait or self.empty():
            must_wait = False
            await self._get_waiting.suspend('QUEUE_GET')

        result = self._get_item()

        if self._put_waiting:
            await self._put_waiting.wake()
        return result

    async def join(self):
        if self._task_count > 0:
            await self._join_waiting.suspend('QUEUE_JOIN')

    async def put(self, item):
        while self.full():
            await self._put_waiting.suspend('QUEUE_PUT')
        self._put_item(item)
        self._task_count += 1
        if self._get_waiting:
            await self._get_waiting.wake()

    def qsize(self):
        return len(self._queue)

    async def task_done(self):
        self._task_count -= 1
        if self._task_count == 0 and self._join_waiting:
            await self._join_waiting.wake()


# UniversalQueue is one of the more interesting, and possibly,
# diabolical features of Curio.  The goal is to provide a Queue that's
# compatible with Curio, Threads, and asyncio using an identical API.
# The underlying operation is based on a combination of non-blocking
# queuing coupled with Futures.   Basically, each underlying operation
# such as get() and put() completes immediately or else a Future
# is created for obtaining the result when it becomes available.
# This works because all of these runtime environments have a mechanism
# for waiting on a Future.   So, the general idea for each queuing
# operation is that you first try the operation.  If it works, you're
# done.  If it doesn't work, you get a Future and you wait on it
# using whatever mechanism the runtime environment uses to do that.

class UniversalQueueBase:
    '''
    A queue for communicating between Curio and external threads,
    including foreign event loops running in different threads.
    '''

    def __init__(self, *, maxsize=0, withfd=False):
        self.maxsize = maxsize

        # The actual queue of items
        self._queue = self._init_internal_queue()

        # A queue of Futures representing getters
        self._getters = deque()

        # A queue of Futures representing putters
        self._putters = deque()

        # Internal synchronization.
        #
        # This is one of the only thread locks that's used inside
        # Curio and used from async code.  It's use here is avoid
        # a race condition on a few attributes.  It is only held
        # briefly, and never in a situation where a blocking operation
        # would take place with the lock held.
        self._mutex = threading.Lock()
        self._all_tasks_done = threading.Condition(self._mutex)
        self._unfinished_tasks = 0

        # Optional socket for monitoring on an external event loop.
        # The fileno() method below returns a socket that becomes
        # readable once an item is available on the queue.
        if withfd:
            self._put_sock, self._get_sock = std_socket.socketpair()
            self._put_sock.setblocking(False)
            self._get_sock.setblocking(False)
        else:
            self._put_sock = self._get_sock = None

    def __del__(self):
        if self._put_sock:
            self._put_sock.close()
            self._get_sock.close()

    def fileno(self):
        assert self._get_sock, "Queue not created with I/O polling enabled"
        return self._get_sock.fileno()

    def _put_send(self):
        if self._put_sock:
            try:
                self._put_sock.send(b'\x01')
            except BlockingIOError:
                pass

    def empty(self):
        return not bool(self._queue)

    def full(self):
        return False

    def qsize(self):
        return len(self._queue)

    def _get_complete(self):
        if self._get_sock:
            try:
                self._get_sock.recv(1)
            except BlockingIOError:
                pass
        self._put_notify()

    def _get(self):
        fut = item = None
        with self._mutex:
            # Critical section never blocks.
            if not self._queue or self._getters:
                fut = Future()
                fut.add_done_callback(lambda f: self._get_complete() if not f.cancelled() else None)
                self._getters.append(fut)
            else:
                item = self._get_item()
                self._get_complete()
        return item, fut

    # Synchronous queue get.
    def get(self):
        item, fut = self._get()
        if fut:
            item = fut.result()
        return item

    # Asynchronous queue get (Curio)
    @awaitable(get)
    async def get(self):
        item, fut = self._get()
        if fut:
            try:
                await _future_wait(fut)
                item = fut.result()
            except CancelledError as e:
                # If we're cancelled, but the future completes successfully anyways,
                # we must arrange for the item to go back onto the queue.  Note:
                # the Curio kernel cancels futures when _future_wait() is cancelled.
                fut.add_done_callback(lambda f: self._put(f.result(), True) if not f.cancelled() else None)
                raise
        return item

    # Asynchronous queue get (Asyncio)
    @asyncioable(get)
    async def get(self):
        item, fut = self._get()
        if fut:
            await asyncio.wrap_future(fut)
            item = fut.result()
        return item

    # Wake any waiting putters
    def _put_notify(self):
        while self._putters:
            putter = self._putters.popleft()
            if not putter.cancelled():
                putter.set_result(None)
                break

    # Put something on the queue or return a Future that must
    # be waited on before retrying.
    def _put(self, item, requeue=False):
        with self._mutex:
            # Critical section never blocks.
            if self.maxsize > 0 and len(self._queue) >= self.maxsize:
                fut = Future()
                self._putters.append(fut)
                return fut

            if requeue:
                # self._queue.appendleft(item)
                self._unget_item(item)
            else:
                self._put_item(item)
                # self._queue.append(item)
                self._unfinished_tasks += 1
                self._put_send()

            # If there are any waiters. Wake one of them and pop a queue item
            while self._getters:
                getter = self._getters.popleft()
                if getter.cancelled():
                    continue
                getter.set_result(self._get_item())  # queue.popleft())
                break

    # Synchronous put. For threads.
    def put(self, item):
        while True:
            fut = self._put(item)
            if not fut:
                break
            fut.result()

    # Asynchronous put. For Curio
    @awaitable(put)
    async def put(self, item):
        while True:
            fut = self._put(item)
            if not fut:
                break
            try:
                await _future_wait(fut)
            except CancelledError:
                # If we're cancelled, but the future completes, it means that the getter alerted
                # a task that space was available, but the alert is lost.  We renotify any waiters.
                fut.add_done_callback(lambda fut: self._put_notify() if not fut.cancelled() else None)
                raise

    # Asynchronous put. For Asyncio.
    @asyncioable(put)
    async def put(self, item):
        while True:
            fut = self._put(item)
            if not fut:
                break
            await asyncio.wrap_future(fut)

    def task_done_sync(self):
        with self._all_tasks_done:
            self._unfinished_tasks -= 1
            assert self._unfinished_tasks >= 0, 'task_done called too many times'
            if self._unfinished_tasks <= 0:
                self._all_tasks_done.notify_all()

    @awaitable(task_done_sync)
    async def task_done(self):
        self.task_done_sync()

    def join_sync(self):
        with self._all_tasks_done:
            while self._unfinished_tasks:
                self._all_tasks_done.wait()

    @awaitable(join_sync)
    async def join(self):
        await workers.block_in_thread(self.join_sync)

    @asyncioable(join)
    async def join(self):
        loop = asyncio.get_event_loop()
        return loop.run_in_executor(None, self.join_sync)


# The following classes implement the low-level queue data structure
# and policies for adding and removing items.

class FIFOImpl:

    def _init_internal_queue(self):
        return deque()

    def _get_item(self):
        return self._queue.popleft()

    def _unget_item(self, item):
        self._queue.appendleft(item)

    def _put_item(self, item):
        self._queue.append(item)


class PriorityImpl:

    def _init_internal_queue(self):
        return []

    def _get_item(self):
        return heappop(self._queue)

    def _put_item(self, item):
        heappush(self._queue, item)

    _unget_item = _put_item


class LIFOImpl:

    def _init_internal_queue(self):
        return []

    def _put_item(self, item):
        self._queue.append(item)

    def _get_item(self):
        return self._queue.pop()

    _unget_item = _put_item

# Concrete Queue implementations

class Queue(QueueBase, FIFOImpl):
    '''
    A First-In First-Out queue for communicating between Curio tasks.
    not safe for communicating between Curio and external
    threads, processes, etc.
    '''
    pass

class PriorityQueue(QueueBase, PriorityImpl):
    '''
    A priority queue for communicating between Curio tasks.
    not safe for communicating between Curio and external
    threads, processes, etc.
    '''
    pass


class LifoQueue(QueueBase, LIFOImpl):
    '''
    A Last-In First-Out queue for communicating between Curio tasks.
    Not safe for communicating between Curio and external
    threads, processes, etc.
    '''
    pass


class UniversalQueue(UniversalQueueBase, FIFOImpl):
    '''
    A FIFO queue for communicating between Curio and external threads,
    including foreign event loops running in different threads.
    '''
    pass





# curio/socket.py
#
# Standin for the standard socket library.  The entire contents of stdlib socket are
# made available here.  However, the socket class is replaced by an async compatible version.
# Certain blocking operations are also replaced by versions safe to use in async.
#

import socket as _socket

__all__ = _socket.__all__

from socket import *
from functools import wraps, partial

from . import workers
from . import io


@wraps(_socket.socket)
def socket(*args, **kwargs):
    return io.Socket(_socket.socket(*args, **kwargs))


@wraps(_socket.socketpair)
def socketpair(*args, **kwargs):
    s1, s2 = _socket.socketpair(*args, **kwargs)
    return io.Socket(s1), io.Socket(s2)


@wraps(_socket.fromfd)
def fromfd(*args, **kwargs):
    return io.Socket(_socket.fromfd(*args, **kwargs))

# Replacements for blocking functions related to domain names and DNS

#@wraps(_socket.create_connection)
#async def create_connection(*args, **kwargs):
#    sock = await workers.run_in_thread(partial(_socket.create_connection, *args, **kwargs))
#    return io.Socket(sock)

async def create_connection(address, timeout=None, source_address=None):
    '''
    Pure async implementation of the socket.create_connection function in standard library
    '''
    host, port = address
    err = None
    for res in await getaddrinfo(host, port, 0, SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        sock = None
        try:
            sock = socket(af, socktype, proto)
            if source_address:
                sock.bind(source_address)
            await sock.connect(sa)
            # Break explicitly a reference cycle
            err = None
            return sock

        except error as _:
            err = _
            if sock is not None:
                await sock.close()

    if err is not None:
        raise err
    else:
        raise OSError("getaddrinfo returns an empty list")

@wraps(_socket.getaddrinfo)
async def getaddrinfo(*args, **kwargs):
    return await workers.run_in_thread(partial(_socket.getaddrinfo, *args, **kwargs))


@wraps(_socket.getfqdn)
async def getfqdn(*args, **kwargs):
    return await workers.run_in_thread(partial(_socket.getfqdn, *args, **kwargs))


@wraps(_socket.gethostbyname)
async def gethostbyname(*args, **kwargs):
    return await workers.run_in_thread(partial(_socket.gethostbyname, *args, **kwargs))


@wraps(_socket.gethostbyname_ex)
async def gethostbyname_ex(*args, **kwargs):
    return await workers.run_in_thread(partial(_socket.gethostbyname_ex, *args, **kwargs))


@wraps(_socket.gethostname)
async def gethostname(*args, **kwargs):
    return await workers.run_in_thread(partial(_socket.gethostname, *args, **kwargs))


@wraps(_socket.gethostbyaddr)
async def gethostbyaddr(*args, **kwargs):
    return await workers.run_in_thread(partial(_socket.gethostbyaddr, *args, **kwargs))


@wraps(_socket.getnameinfo)
async def getnameinfo(*args, **kwargs):
    return await workers.run_in_thread(partial(_socket.getnameinfo, *args, **kwargs))



# curio/sched.py
#
# Task-scheduling primitives. These are used to implement low-level
# scheduling operations needed by higher-level abstractions such
# as Events, Locks, Semaphores, and Queues.

__all__ = [ 'SchedFIFO', 'SchedBarrier' ]

# -- Standard Library

from abc import ABC, abstractmethod
from collections import deque

# -- Curio

from .traps import _scheduler_wait, _scheduler_wake


class SchedBase(ABC):

    def __repr__(self):
        return f'{type(self).__name__}<{len(self)} tasks waiting>'

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def _kernel_suspend(self, task):
        '''
        Suspends a task.  This method *must* return a zero-argument
        callable that removes the just added task from the scheduler
        on cancellation. Called by the kernel.
        '''
        pass

    @abstractmethod
    def _kernel_wake(self, ntasks=1):
        '''
        Wake one or more tasks. Returns a list of the awakened tasks.
        Called by the kernel.
        '''
        pass

    async def suspend(self, reason='SUSPEND'):
        '''
        Suspend the calling task. reason is a string containing
        descriptive text to indicate why (used to set the task state).
        '''
        await _scheduler_wait(self, reason)

    async def wake(self, n=1):
        '''
        Wake one or more suspended tasks.
        '''
        await _scheduler_wake(self, n)


class SchedFIFO(SchedBase):
    '''
    A scheduling FIFO. Tasks sleep and awake in the order of arrival.
    The wake method only awakens a single task. Commonly used to
    implement locks and queues.
    '''
    def __init__(self):
        self._queue = deque()
        self._actual_len = 0

    def __len__(self):
        return self._actual_len

    def _kernel_suspend(self, task):
        # The task is placed inside a 1-item list.  If cancelled, the
        # task is replaced by None, but the list remains on the queue
        # until later pop operations discard it
        item = [task]
        self._queue.append(item)
        self._actual_len += 1

        def remove():
            item[0] = None
            self._actual_len -= 1
        return remove

    def _kernel_wake(self, ntasks=1):
        tasks = []
        while ntasks > 0:
            task, = self._queue.popleft()
            if task:
                tasks.append(task)
                ntasks -= 1
        self._actual_len -= len(tasks)
        return tasks

class SchedBarrier(SchedBase):
    '''
    A scheduling barrier.  Sleeping tasks are collected into a set.
    Waking makes all of the blocked tasks reawaken at the same time.
    Commonly used to implement Event and join().
    '''
    def __init__(self):
        self._tasks = set()

    def __len__(self):
        return len(self._tasks)

    def _kernel_suspend(self, task):
        self._tasks.add(task)
        return lambda: self._tasks.remove(task)

    def _kernel_wake(self, ntasks=1):
        if ntasks == len(self._tasks):
            result = list(self._tasks)
            self._tasks.clear()
        else:
            result = [self._tasks.pop() for _ in range(ntasks)]
        return result

    async def wake(self, n=None):
        '''
        Wake all or a specified number of tasks.
        '''
        n = len(self._tasks) if n is None else n
        await _scheduler_wake(self, n)




# curio/errors.py
#
# Curio specific exceptions

__all__ = [
    'CurioError', 'CancelledError', 'TaskTimeout', 'TaskError',
    'SyncIOError', 'ResourceBusy',
    'ReadResourceBusy', 'WriteResourceBusy',
    'TimeoutCancellationError', 'UncaughtTimeoutError',
    'TaskCancelled', 'AsyncOnlyError',
]


class CurioError(Exception):
    '''
    Base class for all non-cancellation Curio-related exceptions
    '''


class CancelledError(BaseException):
    '''
    Base class for all task-cancellation related exceptions
    '''


class TaskCancelled(CancelledError):
    '''
    Exception raised as a result of a task being directly cancelled.
    '''


class TimeoutCancellationError(CancelledError):
    '''
    Exception raised if task is being cancelled due to a timeout, but
    it's not the inner-most timeout in effect.
    '''


class TaskTimeout(CancelledError):
    '''
    Exception raised if task is cancelled due to timeout.
    '''


class UncaughtTimeoutError(CurioError):
    '''
    Raised if a TaskTimeout exception escapes a timeout handling
    block and is unexpectedly caught by an outer timeout handler.
    '''


class TaskError(CurioError):
    '''
    Raised if a task launched via spawn() or similar function
    terminated due to an exception.  This is a chained exception.
    The __cause__ attribute contains the actual exception that
    occurred in the task.
    '''


class SyncIOError(CurioError):
    '''
    Raised if a task attempts to perform a synchronous I/O operation
    on an object that only supports asynchronous I/O.
    '''


class AsyncOnlyError(CurioError):
    '''
    Raised by the AWAIT() function if its applied to code not
    properly running in an async-thread.
    '''


class ResourceBusy(CurioError):
    '''
    Raised by I/O related functions if an operation is requested,
    but the resource is already busy performing the same operation
    on behalf of another task.
    '''


class ReadResourceBusy(ResourceBusy):
    pass


class WriteResourceBusy(ResourceBusy):
    pass





# timequeue.py
#
# A Discussion About Time.
#
# Internally, Curio must manage time for two different reasons,
# sleeping and for timeouts.  Aside from toy examples, most real-world
# code isn't going to sit around making a lot of sleep() calls.
# Instead the more common use is timeouts.  Timeouts are kind of
# interesting though--when a timeout is set, there is typically an
# expectation that it will probably NOT occur.  The expiration of a
# timeout is an exceptional event. Most of the time, a timeout will be
# cancelled before it is allowed to expire.
#
# This presents an interesting implementation challenge for managing
# time.  It is most common to see time managed by sorting the
# expiration times in some way. For example, placing them in a sorted
# list, or ordering them on a heap in a priority queue.  Although
# operations on these kinds of data structures can be managed in O(log N)
# steps, they might not be necessary at all if you make some slightly
# different assumptions about time management.
#
# The queue implementation here is based on the idea that expiration
# times in the distant future don't need to be precisely sorted.
# Instead, you can merely drop expiration times in a dict with the
# hope that they'll be cancelled later.  Manipulating a dict in this
# case is O(1)--meaning that is extremely cheap to setup and teardown
# a timeout that never occurs.  For timeouts in the near future, they
# can still be sorted using a priority queue in the usual way.

import heapq

class TimeQueue:
    cutoff = 1.0         # Threshhold for near/far events (seconds)
    def __init__(self):
        self.near = [ ]
        self.far = { }
        self.near_deadline = 0
        self.far_min_deadline = float('inf')

    def _far_to_near(self):
        '''
        Move items from the far queue to the near queue (if any).
        '''
        removed = []
        min_deadline = float('inf')
        for item, expires in self.far.items():
            if expires < self.near_deadline:
                self.push(item, expires)
                removed.append(item)
            elif expires < min_deadline:
                min_deadline = expires
        for item in removed:
            del self.far[item]
        self.far_min_deadline = min_deadline

    def next_deadline(self, current_clock):
        '''
        Returns the number of seconds to delay until the next deadline
        expires.  current_clock is the current value of the clock.
        Returns None if there are no pending deadlines.
        '''
        self.near_deadline = current_clock + self.cutoff
        if self.near_deadline > self.far_min_deadline:
            self._far_to_near()

        if self.near:
            delta = self.near[0][0] - current_clock
            return delta if delta > 0 else 0

        # There are no near deadlines. Use the closest far deadline
        if self.far:
            delta = self.far_min_deadline - current_clock
            return delta if delta > 0 else 0

        # There are no sleeping tasks of any kind.
        return None

    def push(self, item, expires):
        '''
        Push a new item onto the time queue.
        '''
        # If the expiration time is closer than the current near deadline,
        # it gets pushed onto a heap in order to preserve order
        if expires <= self.near_deadline:
            heapq.heappush(self.near, (expires, item))
        else:
            # Otherwise the item gets put into a dict for far-in-future handling
            if item not in self.far or self.far[item] > expires:
                self.far[item] = expires
            if expires < self.far_min_deadline:
                self.far_min_deadline = expires

    def expired(self, deadline):
        '''
        An iterator that returns all items that have expired up to a given deadline
        '''
        near = self.near
        if deadline >= self.far_min_deadline:
            self.near_deadline = deadline + self.cutoff
            self._far_to_near()

        while near and near[0][0] <= deadline:
            yield  heapq.heappop(near)

    def cancel(self, item, expires):
        '''
        Cancel a time event. The combination of (item, expires) should
        match a prior push() operation (but if not, it's ignored).
        '''
        self.far.pop(item, None)



# channel.py
#
# Support for a message passing channel that can send bytes or pickled
# Python objects on a stream.  Compatible with the Connection class in the
# multiprocessing module, but rewritten for a purely asynchronous runtime.

__all__ = ['Channel']

# -- Standard Library

import os
import pickle
import struct
import hmac
import multiprocessing.connection as mpc
import logging

log = logging.getLogger(__name__)

# -- Curio

from . import socket
from .errors import CurioError, TaskTimeout
from .io import StreamBase, FileStream
from . import thread
from .time import timeout_after, sleep

# Authentication parameters (copied from multiprocessing)

AUTH_MESSAGE_LENGTH = mpc.MESSAGE_LENGTH    # 20
CHALLENGE = mpc.CHALLENGE                   # b'#CHALLENGE#'
WELCOME = mpc.WELCOME                       # b'#WELCOME#'
FAILURE = mpc.FAILURE                       # b'#FAILURE#'



class ConnectionError(CurioError):
    pass


class AuthenticationError(ConnectionError):
    pass


class Connection(object):
    '''
    A communication channel for sending size-prefixed messages of bytes
    or pickled Python objects.  Must be passed a pair of reader/writer
    streams for performing the underlying communication.
    '''

    def __init__(self, reader, writer):
        assert isinstance(reader, StreamBase) and isinstance(writer, StreamBase)
        self._reader = reader
        self._writer = writer

    @classmethod
    def from_Connection(cls, conn):
        '''
        Creates a channel from a multiprocessing Connection. Note: The
        multiprocessing connection is detached by having its handle set to None.

        This method can be used to make curio talk over Pipes as created by
        multiprocessing.  For example:

              p1, p2 = multiprocessing.Pipe()
              p1 = Connection.from_Connection(p1)
              p2 = Connection.from_Connection(p2)

        '''
        assert isinstance(conn, mpc._ConnectionBase)
        reader = FileStream(open(conn._handle, 'rb', buffering=0))
        writer = FileStream(open(conn._handle, 'wb', buffering=0, closefd=False))
        conn._handle = None
        return cls(reader, writer)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()

    def __enter__(self):
        return thread.AWAIT(self.__aenter__())

    def __exit__(self, *args):
        return thread.AWAIT(self.__aexit__(*args))

    async def close(self):
        await self._reader.close()
        if self._reader != self._writer:
            await self._writer.close()

    async def send_bytes(self, buf, offset=0, size=None):
        '''
        Send a buffer of bytes as a single message
        '''
        m = memoryview(buf)
        if m.itemsize > 1:
            m = memoryview(bytes(m))
        n = len(m)
        if offset < 0:
            raise ValueError("offset is negative")
        if n < offset:
            raise ValueError("buffer length < offset")
        if size is None:
            size = n - offset
        elif size < 0:
            raise ValueError("size is negative")
        elif offset + size > n:
            raise ValueError("buffer length < offset + size")

        header = struct.pack('!i', size)
        if size >= 16384:
            await self._writer.write(header)
            await self._writer.write(m[offset:offset + size])
        else:
            msg = header + bytes(m[offset:offset + size])
            await self._writer.write(msg)
        return size

    async def recv_bytes(self, maxlength=None):
        '''
        Receive a message of bytes as a single message.
        '''
        header = await self._reader.read_exactly(4)
        size, = struct.unpack('!i', header)
        if maxlength and size > maxlength:
            raise IOError("Message too large")
        msg = await self._reader.read_exactly(size)
        return msg

    async def recv_bytes_into(self, buf, offset=0):
        '''
        Receive bytes into a writable memory buffer.  The buffer must be large enough to
        hold the message.  The number of bytes received in the message is returned.
        '''
        header = await self._reader.read_exactly(4)
        size, = struct.unpack('!i', header)
        with memoryview(buf).cast('B') as m:
            if size > (len(m) - offset):
                # Message is too large to fit in allotted space
                # Drain the I/O and raise an error
                while size > 0:
                    data = await self._reader.read(size)
                    if not data:
                        break
                    size -= len(data)
                raise IOError('Message is too large to fit')
            nread = await self._reader.readinto(m[offset:offset+size])
            if nread != size:
                raise EOFError('Expected end of data')
            return nread

    async def send(self, obj):
        '''
        Send an arbitrary Python object. Uses pickle to serialize.
        '''
        await self.send_bytes(pickle.dumps(obj, pickle.HIGHEST_PROTOCOL))

    async def recv(self):
        '''
        Receive a Python object. Uses pickle to unserialize.
        '''
        msg = await self.recv_bytes()
        return pickle.loads(msg)

    async def _deliver_challenge(self, authkey):
        message = os.urandom(AUTH_MESSAGE_LENGTH)
        await self.send_bytes(CHALLENGE + message)
        digest = hmac.new(authkey, message, 'md5').digest()
        response = await self.recv_bytes(maxlength=256)
        if response == digest:
            await self.send_bytes(WELCOME)
        else:
            await self.send_bytes(FAILURE)
            raise AuthenticationError('digest received was wrong')

    async def _answer_challenge(self, authkey):
        message = await self.recv_bytes(maxlength=256)
        assert message[:len(CHALLENGE)] == CHALLENGE, f'message = {message!r}'
        message = message[len(CHALLENGE):]
        digest = hmac.new(authkey, message, 'md5').digest()
        await self.send_bytes(digest)
        response = await self.recv_bytes(maxlength=256)

        if response != WELCOME:
            raise AuthenticationError('digest sent was rejected')

    async def authenticate_server(self, authkey):
        await self._deliver_challenge(authkey)
        await self._answer_challenge(authkey)

    async def authenticate_client(self, authkey):
        await self._answer_challenge(authkey)
        await self._deliver_challenge(authkey)

class Channel(object):
    def __init__(self, address, family=socket.AF_INET, check_address=None):
        self.address = address
        self.family = family
        self.sock = None
        if check_address:
            self.check_address = check_address

    def __repr__(self):
        return f'Channel({self.address!r}, {self.family!r})'

    async def __aenter__(self):
        return self

    async def __aexit__(self, ty, val, tb):
        await self.close()

    def __getstate__(self):
        return (self.address, self.family)

    def __setstate__(self, state):
        self.address, self.family = state
        self.sock = None

    def bind(self):
        self.sock = socket.socket(self.family, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        self.sock.bind(self.address)
        self.sock.listen(5)
        self.address = self.sock.getsockname()

    def check_address(self, addr):
        return True

    async def accept(self, *, authkey=None):
        if self.sock is None:
            self.bind()

        while True:
            client, addr = await self.sock.accept()
            if not self.check_address(addr):
                log.warning('Channel connection from %s rejected', addr)
                await client.close()
                del client
                continue

            client_stream = client.as_stream()
            c = Connection(client_stream, client_stream)
            c.address = addr
            try:
                async with timeout_after(1):
                    if authkey:
                        await c.authenticate_server(authkey)
                break
            except (TaskTimeout, AuthenticationError, EOFError):
                log.warning('Channel connection from %s failed', addr, exc_info=True)
                await c.close()
                del c
                del client_stream
                del client
        return c

    async def connect(self, *, authkey=None):
        sock = socket.socket(self.family, socket.SOCK_STREAM)
        await sock.connect(self.address)
        sock_stream = sock.as_stream()
        c = Connection(sock_stream, sock_stream)
        try:
            async with timeout_after(1):
                if authkey:
                    await c.authenticate_client(authkey)
            return c
        except TaskTimeout:
            log.warning('Channel connection to %s timed out', self.address)
            await c.close()
            del c
            del sock_stream
            # Note: Raising an OSError.
            raise TimeoutError("Connection timed out")

    async def close(self):
        if self.sock:
            await self.sock.close()
        self.sock = None



# curio/thread.py
#
# Support for threads implemented on top of the Curio kernel.
#
# Theory of operation:
# --------------------
# Curio has the ability to safely wait for Futures as defined
# in the concurrent.futures module.  A notable feature of coroutines
# is that when called, their evaluation is delayed--instead you get
# a "coroutine" object that must be executed by a kernel or event loop.
#
# A so-called "async thread" uses both of these features together to
# set up an execution pathway for allowing threads to execute
# coroutines.  For each thread (a real thread--created by the
# threading module), a backing coroutine is created in Curio.  This
# backing coroutine runs on top of the Curio kernel and constantly
# monitors a Future for an incoming request.  This request is expected
# to contain an unevaluated coroutine. The unevaluated coroutine is
# evaluated on behalf of the thread by the backing coroutine.  Any
# result is the communicated back to the thread which is waiting
# for it on an Event.
#
# The mechanism for making a request within a thread is the AWAIT
# function.  Specifically, a call like this:
#
#       result = AWAIT(coro, *args, **kwargs)
#
# Makes the thread's backing coroutine execute the following:
#
#       result = await coro(*args, **kwargs)
#
# From the standpoint of the thread, it appears to be executing a
# normal synchronous call.
#
# Here is a picture diagram of the parts
#
#    ________          ___________          _________
#   |        | await  |           | Future |         |
#   | Curio  |<-------| backing   |<-------| Thread  |
#   | Kernel |------->| coroutine |------->|         |
#   |________| result |___________| Event  |_________|
#

__all__ = [ 'AWAIT', 'spawn_thread' ]

# -- Standard Library

import threading
from concurrent.futures import Future
from functools import wraps
from inspect import iscoroutine, isgenerator
from contextlib import contextmanager
import logging

log = logging.getLogger(__name__)

# -- Curio

from . import sync
from . import queue
from .task import spawn, disable_cancellation, check_cancellation, set_cancellation
from .traps import _future_wait
from . import errors
from . import meta

_locals = threading.local()

class AsyncThread(object):

    def __init__(self, target=None, args=(), kwargs={}, daemon=False):
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.daemon = daemon

        # The following attributes are provided to make a thread mimic a Task
        self.terminated = False
        self.cancelled = False
        self.taskgroup = None
        self.joined = False

        # This future is used by a thread to make a request to Curio
        self._request = Future()

        # This event is used to communicate completion of the request
        self._done_evt = threading.Event()

        # Event used to signal thread termination
        self._terminate_evt = sync.UniversalEvent()

        # Information about the coroutine being executed by the thread
        self._coro = None
        self._coro_result = None
        self._coro_exc = None

        # Final values produced by the thread before termination
        self._final_value = None
        self._final_exc = None

        # A reference to the associated thread (from threading module)
        self._thread = None

        # A reference to the associated backing task
        self._task = None

    async def _coro_runner(self):
        while True:
            # Wait for a hand-off
            await disable_cancellation(_future_wait(self._request))
            self._coro = self._request.result()
            self._request = Future()

            # If no coroutine, we're shutting down
            if not self._coro:
                break

            # Run the the coroutine
            try:
                self._coro_result = await self._coro
                self._coro_exc = None
            except BaseException as e:
                self._coro_result = None
                self._coro_exc = e

            # Hand it back to the thread
            self._coro = None
            self._done_evt.set()

        if self.taskgroup:
            await self.taskgroup._task_done(self)
            self.joined = True
        await self._terminate_evt.set()

    def _func_runner(self):
        _locals.thread = self
        try:
            self._final_result = self.target(*self.args, **self.kwargs)
            self._final_exc = None
        except BaseException as e:
            self._final_result = None
            self._final_exc = e
            if not isinstance(e, errors.CancelledError):
                log.warning("Unexpected exception in cancelled async thread", exc_info=True)

        finally:
            self._request.set_result(None)

    async def start(self):
        if self.target is None:
            raise RuntimeError("Async thread must be given a target")

        # Launch the backing coroutine
        self._task = await spawn(self._coro_runner, daemon=True)

        # Launch the thread itself
        self._thread = threading.Thread(target=self._func_runner)
        self._thread.start()

    def AWAIT(self, coro):
        self._request.set_result(coro)
        self._done_evt.wait()
        self._done_evt.clear()

        if self._coro_exc:
            raise self._coro_exc
        else:
            return self._coro_result

    async def join(self):
        await self.wait()
        self.joined = True
        if self.taskgroup:
            self.taskgroup._task_discard(self)

        if self._final_exc:
            raise errors.TaskError() from self._final_exc
        else:
            return self._final_result

    async def wait(self):
        await self._terminate_evt.wait()
        self.terminated = True

    @property
    def result(self):
        if not self._terminate_evt.is_set():
            raise RuntimeError('Thread not terminated')
        if self._final_exc:
            raise self._final_exc
        else:
            return self._final_result

    @property
    def exception(self):
        if not self._terminate_evt.is_set():
            raise RuntimeError('Thread not terminated')
        return self._final_exc

    async def cancel(self, *, exc=errors.TaskCancelled, blocking=True):
        self.cancelled = True
        await self._task.cancel(exc=exc, blocking=blocking)
        if blocking:
            await self.wait()

    @property
    def id(self):
        return self._task.id

    @property
    def state(self):
        return self._task.state

def AWAIT(coro, *args, **kwargs):
    '''
    Await for a coroutine in an asynchronous thread.  If coro is
    not a proper coroutine, this function acts a no-op, returning coro.
    '''
    # If the coro is a callable and it's identifiable as a coroutine function,
    # wrap it inside a coroutine and pass that.
    if callable(coro):
        if meta.iscoroutinefunction(coro) and hasattr(_locals, 'thread'):
            async def _coro(coro):
                return await coro(*args, **kwargs)
            coro = _coro(coro)
        else:
            coro = coro(*args, **kwargs)

    if iscoroutine(coro) or isgenerator(coro):
        if hasattr(_locals, 'thread'):
            return _locals.thread.AWAIT(coro)
        else:
            # Thought: Do we try to promote the calling thread into an
            # "async" thread automatically?  Would require a running
            # kernel. Would require a task dedicated to spawning the
            # coro runner.  Would require shutdown.  Maybe a context
            # manager?
            raise errors.AsyncOnlyError('Must be used as async')
    else:
        return coro

def spawn_thread(func, *args, daemon=False):
    '''
    Launch an async thread.  This mimicks the way a task is normally spawned. For
    example:

         t = await spawn_thread(func, arg1, arg2)
         ...
         await t.join()
    '''
    if iscoroutine(func) or meta.iscoroutinefunction(func):
          raise TypeError("spawn_thread() can't be used on coroutines")

    async def runner(args, daemon):
        t = AsyncThread(func, args=args, daemon=daemon)
        await t.start()
        return t

    return runner(args, daemon)

def is_async_thread():
    '''
    Returns True if current thread is an async thread.
    '''
    return hasattr(_locals, 'thread')



# curio/debug.py
#
# Task debugging tools

__all__ = [ 'longblock', 'schedtrace', 'traptrace', 'logcrash' ]

import time
import logging
log = logging.getLogger(__name__)

# -- Curio

from .kernel import Activation
from .errors import TaskCancelled

class DebugBase(Activation):
    def __init__(self, *, level=logging.INFO, log=log, filter=None, **kwargs):
        self.level = level
        self.filter = filter
        self.log = log

    def check_filter(self, task):
        if self.filter and task.name not in self.filter:
            return False
        return True

class longblock(DebugBase):
    '''
    Report warnings for tasks that block the event loop for a long duration.
    '''
    def __init__(self, *, max_time=0.05, level=logging.WARNING, **kwargs):
        super().__init__(level=level, **kwargs)
        self.max_time = max_time

    def running(self, task):
        if self.check_filter(task):
            self.start = time.monotonic()

    def suspended(self, task, trap):
        if self.check_filter(task):
            duration = time.monotonic() - self.start
            if duration > self.max_time:
                self.log.log(self.level, '%r ran for %s seconds', task, duration)

class logcrash(DebugBase):
    '''
    Report tasks that crash with an uncaught exception
    '''
    def __init__(self, level=logging.ERROR, **kwargs):
        super().__init__(level=level, **kwargs)

    def suspended(self, task, trap):
        if task.terminated and self.check_filter(task):
            if task.exception and not isinstance(task.exception, (StopIteration, TaskCancelled, KeyboardInterrupt, SystemExit)):
                self.log.log(self.level, '%r crashed', task, exc_info=task.exception)

class schedtrace(DebugBase):
    '''
    Report when tasks run
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def created(self, task):
        if self.check_filter(task):
            self.log.log(self.level, 'CREATE:%f:%r', time.time(), task)

    def running(self, task):
        if self.check_filter(task):
            self.log.log(self.level, 'RUN:%f:%r', time.time(), task)

    def suspended(self, task, trap):
        if self.check_filter(task):
            self.log.log(self.level, 'SUSPEND:%f:%r', time.time(), task)

    def terminated(self, task):
        if self.check_filter(task):
            self.log.log(self.level, 'TERMINATED:%f:%r', time.time(), task)

class traptrace(schedtrace):
    '''
    Report traps executed
    '''
    def suspended(self, task, trap):
        if self.check_filter(task):
            if trap:
                self.log.log(self.level, 'TRAP:%r', trap)
            super().suspended(task, trap)

def _create_debuggers(debug):
    '''
    Create debugger objects.  Called by the kernel to instantiate the objects.
    '''
    if debug is True:
        # Set a default set of debuggers
        debug = [ schedtrace ]

    elif not isinstance(debug, (list, tuple)):
        debug = [ debug ]

    # Create instances
    debug = [ (d() if (isinstance(d, type) and issubclass(d, DebugBase)) else d)
              for d in debug ]
    return debug









# curio/__init__.py

__version__ = '1.6'

from .errors import *
from .queue import *
from .task import *
from .time import *
from .kernel import *
from .sync import *
from .workers import *
from .network import *
from .file import *
from .channel import *
from .thread import *

__all__ = [*errors.__all__,
           *queue.__all__,
           *task.__all__,
           *time.__all__,
           *kernel.__all__,
           *sync.__all__,
           *workers.__all__,
           *network.__all__,
           *file.__all__,
           *channel.__all__,
           *thread.__all__,
           ]



# curio/sync.py
#
# Implementation of common task synchronization primitives such as
# events, locks, semaphores, and condition variables. These primitives
# are only safe to use within Curio--they are not thread safe
# unless otherwise indicated.
#
# The general implementation strategy is based on task scheduling.
# For example, if a task needs to wait on a lock, it goes to sleep on
# a queue.  When a task releases a lock, it wakes a sleeping task.
# Task scheduling is provided by the SchedFIFO and SchedBarrier classes in sched.py

__all__ = ['Event', 'UniversalEvent', 'Lock', 'RLock', 'Semaphore', 'Condition', 'Result', 'UniversalResult' ]

# -- Standard library

import threading
from concurrent.futures import Future
import asyncio

# -- Curio

from .sched import SchedFIFO, SchedBarrier
from . import workers
from .task import current_task
from .meta import awaitable, asyncioable, iscoroutinefunction
from . import thread
from .traps import _future_wait

class Event(object):

    def __init__(self):
        self._set = False
        self._waiting = SchedBarrier()

    def __repr__(self):
        res = super().__repr__()
        extra = 'set' if self._set else 'unset'
        return f'<{res[1:-1]} [{extra},waiters:{len(self._waiting)}]>'

    def is_set(self):
        return self._set

    def clear(self):
        self._set = False

    async def wait(self):
        if self._set:
            return
        await self._waiting.suspend('EVENT_WAIT')

    async def set(self):
        self._set = True
        await self._waiting.wake()

class UniversalEvent(object):
    '''
    An event that's safe to use from Curio and threads.
    '''
    def __init__(self):
        self._set = False
        self._lock = threading.Lock()
        self._waiting = set()

    def __repr__(self):
        res = super().__repr__()
        extra = 'set' if self._set else 'unset'
        return f'<{res[1:-1]} [{extra}]>'

    def is_set(self):
        return self._set

    def clear(self):
        with self._lock:
            self._set = False

    def wait(self):
        with self._lock:
            if self._set:
                return
            fut = Future()
            self._waiting.add(fut)
        try:
            fut.result()
        finally:
            self._waiting.discard(fut)

    @awaitable(wait)
    async def wait(self):
        with self._lock:
            if self._set:
                return
            fut = Future()
            self._waiting.add(fut)
        try:
            await _future_wait(fut)
        finally:
            self._waiting.discard(fut)

    @asyncioable(wait)
    async def wait(self):
        with self._lock:
            if self._set:
                return
            fut = Future()
            self._waiting.add(fut)
        try:
            await asyncio.wrap_future(fut)
        finally:
            self._waiting.discard(fut)

    def _unblock_waiters(self):
        # Caution: Should only be called with self._lock held.
        now_waiting, self._waiting = self._waiting, set()
        for fut in now_waiting:
            if not fut.done():
                fut.set_result(True)
            
    def set(self):
        with self._lock:
            if self._set:
                return
            self._set = True
            self._unblock_waiters()

    @awaitable(set)
    async def set(self):
        with self._lock:
            if self._set:
                return
            self._set = True
            self._unblock_waiters()        

    @asyncioable(set)
    async def set(self):
        with self._lock:
            if self._set:
                return
            self._set = True
            self._unblock_waiters()                

# Base class for all synchronization primitives that operate as context managers.

class _LockBase(object):

    async def __aenter__(self):
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.release()

    def __enter__(self):
        return thread.AWAIT(self.__aenter__())

    def __exit__(self, *args):
        return thread.AWAIT(self.__aexit__(*args))

class Lock(_LockBase):

    def __init__(self):
        self._acquired = False
        self._waiting = SchedFIFO()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return f'<{res[1:-1]} [{extra},waiters:{len(self._waiting)}]>'

    async def acquire(self):
        if self._acquired:
            await self._waiting.suspend('LOCK_ACQUIRE')
        self._acquired = True
        return True

    async def release(self):
        assert self._acquired, 'Lock not acquired'
        if self._waiting:
            await self._waiting.wake()
        else:
            self._acquired = False

    def locked(self):
        return self._acquired

class RLock(_LockBase):

    def __init__(self):
        self._lock = Lock()
        self._owner = None
        self._count = 0

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return f'<{res[1:-1]} [{extra},recursion:{self._count}]>'

    async def acquire(self):

        me = await current_task()

        if self._owner is not me:
            await self._lock.acquire()
            self._owner = me

        self._count += 1
        return True

    async def release(self):
        if not self.locked():
            raise RuntimeError('RLock is not locked')
        if not await current_task() is self._owner:
            raise RuntimeError('RLock can only be released by the owner')
        self._count -= 1
        if self._count == 0:
            await self._lock.release()
            self._owner = None

    def locked(self):
        return self._count > 0


class Semaphore(_LockBase):

    def __init__(self, value=1):
        self._value = value
        self._waiting = SchedFIFO()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return f'<{res[1:-1]} [{extra},value:{self._value},waiters:{len(self._waiting)}]>'

    @property
    def value(self):
        return self._value

    async def acquire(self):
        if self._value <= 0:
            await self._waiting.suspend('SEMA_ACQUIRE')
        else:
            self._value -= 1
        return True

    async def release(self):
        if self._waiting:
            await self._waiting.wake()
        else:
            self._value += 1

    def locked(self):
        return self._value == 0


class Condition(_LockBase):

    def __init__(self, lock=None):
        if lock is None:
            self._lock = Lock()
        else:
            self._lock = lock
        self._waiting = SchedFIFO()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return f'<{res[1:-1]} [{extra},waiters:{len(self._waiting)}]>'

    def locked(self):
        return self._lock.locked()

    async def acquire(self):
        await self._lock.acquire()

    async def release(self):
        await self._lock.release()

    async def wait(self):
        if not self.locked():
            raise RuntimeError("Can't wait on unacquired lock")
        await self.release()
        try:
            await self._waiting.suspend('COND_WAIT')
        finally:
            await self.acquire()

    async def wait_for(self, predicate):
        while True:
            result = predicate()
            if result:
                return result
            await self.wait()

    async def notify(self, n=1):
        if not self.locked():
            raise RuntimeError("Can't notify on unacquired lock")
        await self._waiting.wake(n)

    async def notify_all(self):
        await self.notify(len(self._waiting))

class Result:
    def __init__(self):
        self._evt = Event()
        self._value = None
        self._exc = None

    def __repr__(self):
        res = super().__repr__()        
        if self._evt.is_set():
            return f'<{res[1:-1]}, value={self._value!r}, exc={self._exc!r}>'
        else:
            return f'<{res[1:-1]}, not set>'
        
        status = "set" if self.is_set() else "not set"
        return f'<Result status={status}>'
    
    def is_set(self):
        return self._evt.is_set()

    async def unwrap(self):
        await self._evt.wait()
        if self._exc:
            raise self._exc from None
        else:
            return self._value

    async def set_value(self, value):
        self._value = value
        await self._evt.set()
        
    async def set_exception(self, exc):
        self._exc = exc
        await self._evt.set()

class UniversalResult:
    
    def __init__(self):
        self._evt = UniversalEvent()
        self._value = None
        self._exc = None

    def __repr__(self):
        res = super().__repr__()
        if self._evt.is_set():
            return f'<{res[1:-1]}, value={self._value!r}, exc={self._exc!r}>'
        else:
            return f'<{res[1:-1]}, not set>'

    def is_set(self):
        return self._evt.is_set()

    def _return_result(self):
        if self._exc:
            raise self._exc from None
        else:
            return self._value

    def unwrap(self):
        self._evt.wait()
        return self._return_result()

    @awaitable(unwrap)
    async def unwrap(self):
        await self._evt.wait()
        return self._return_result()

    def set_value(self, value):
        self._value = value
        self._evt.set()

    @awaitable(set_value)
    async def set_value(self, value):
        self._value = value
        await self._evt.set()

    def set_exception(self, exc):
        self._exc = exc
        self._evt.set()

    @awaitable(set_exception)
    async def set_exception(self, exc):
        self._exc = exc
        await self._evt.set()





# traps.py
#
# Curio programs execute under the supervision of a
# kernel. Communication with the kernel takes place via a "trap"
# involving the yield statement.  Traps represent internel kernel
# procedures.  Direct use of the functions defined here is allowed
# when making new kinds of Curio primitives, but if you're trying to
# solve a higher level problem, there is probably a higher-level
# interface that is easier to use (e.g., Socket, File, Queue, etc.).
# ----------------------------------------------------------------------

__all__ = [
    '_read_wait', '_write_wait', '_future_wait', '_sleep', '_spawn',
    '_cancel_task', '_scheduler_wait', '_scheduler_wake',
    '_get_kernel', '_get_current', '_set_timeout', '_unset_timeout',
    '_clock', '_io_waiting', '_io_release',
    ]

# -- Standard library

from types import coroutine
from selectors import EVENT_READ, EVENT_WRITE

# -- Curio

from . import errors

# This is the only entry point to the Curio kernel and the
# only place where the @types.coroutine decorator is used.
@coroutine
def _kernel_trap(*request):
    result = yield request
    if isinstance(result, BaseException):
        raise result
    else:
        return result

# Higher-level trap functions that make use of async/await
async def _read_wait(fileobj):
    '''
    Wait until reading can be performed.  If another task is waiting
    on the same file, a ResourceBusy exception is raised.
    '''
    return await _kernel_trap('trap_io', fileobj, EVENT_READ, 'READ_WAIT')

async def _write_wait(fileobj):
    '''
    Wait until writing can be performed. If another task is waiting
    to write on the same file, a ResourceBusy exception is raised.
    '''
    return await _kernel_trap('trap_io', fileobj, EVENT_WRITE, 'WRITE_WAIT')

async def _io_release(fileobj):
    '''
    Release kernel resources associated with a file
    '''
    return await _kernel_trap('trap_io_release', fileobj)

async def _io_waiting(fileobj):
    '''
    Return a tuple (rtask, wtask) of tasks currently blocked waiting
    for I/O on fileobj.
    '''
    return await _kernel_trap('trap_io_waiting', fileobj)

async def _future_wait(future, event=None):
    '''
    Wait for the result of a Future to be ready.
    '''
    return await _kernel_trap('trap_future_wait', future, event)

async def _sleep(clock):
    '''
    Sleep until the monotonic clock reaches the specified clock value.
    If clock is 0, forces the current task to yield to the next task (if any).
    '''
    return await _kernel_trap('trap_sleep', clock)

async def _spawn(coro):
    '''
    Create a new task. Returns the resulting Task object.
    '''
    return await _kernel_trap('trap_spawn', coro)

async def _cancel_task(task, exc=errors.TaskCancelled, val=None):
    '''
    Cancel a task. Causes a CancelledError exception to raise in the task.
    Set the exc and val arguments to change the exception.
    '''
    return await _kernel_trap('trap_cancel_task', task, exc, val)

async def _scheduler_wait(sched, state):
    '''
    Put the task to sleep on a scheduler primitive.
    '''
    return await _kernel_trap('trap_sched_wait', sched, state)

async def _scheduler_wake(sched, n=1):
    '''
    Reschedule one or more tasks waiting on a scheduler primitive.
    '''
    return await _kernel_trap('trap_sched_wake', sched, n)

async def _get_kernel():
    '''
    Get the kernel executing the task.
    '''
    return await _kernel_trap('trap_get_kernel')

async def _get_current():
    '''
    Get the currently executing task
    '''
    return await _kernel_trap('trap_get_current')

async def _set_timeout(clock):
    '''
    Set a timeout for the current task that occurs at the specified clock value.
    Setting a clock of None clears any previous timeout.
    '''
    return await _kernel_trap('trap_set_timeout', clock)

async def _unset_timeout(previous):
    '''
    Restore the previous timeout for the current task.
    '''
    return await _kernel_trap('trap_unset_timeout', previous)

async def _clock():
    '''
    Return the value of the kernel clock
    '''
    return await _kernel_trap('trap_clock')



import ast
import curio
import curio.monitor
import code
import inspect
import sys
import types
import warnings
import threading
import signal
import os

assert (sys.version_info.major >= 3 and sys.version_info.minor >= 8), "console requires Python 3.8+"

class CurioIOInteractiveConsole(code.InteractiveConsole):

    def __init__(self, locals):
        super().__init__(locals)
        self.compile.compiler.flags |= ast.PyCF_ALLOW_TOP_LEVEL_AWAIT
        self.requests = curio.UniversalQueue()
        self.response = curio.UniversalQueue()

    def runcode(self, code):
        # This coroutine is handed from the thread running the REPL to the
        # task runner in the main thread.
        async def run_it():
            func = types.FunctionType(code, self.locals)
            try:
                # We restore the default REPL signal handler for running normal code
                hand = signal.signal(signal.SIGINT, signal.default_int_handler)
                try:
                    coro = func()
                finally:
                    signal.signal(signal.SIGINT, hand)
            except BaseException as ex:
                await self.response.put((None, ex))
                return
            if not inspect.iscoroutine(coro):
                await self.response.put((coro, None))
                return

            # For a coroutine... We're going to try and do some magic to intercept
            # Control-C in an Event/Task.
            async def watch_ctrl_c(evt, repl_task):
                await evt.wait()
                await repl_task.cancel()
            evt = curio.UniversalEvent()
            try:
                hand = signal.signal(signal.SIGINT, lambda signo, frame: evt.set())
                repl_task = await curio.spawn(coro)
                watch_task = await curio.spawn(watch_ctrl_c, evt, repl_task)
                try:
                    result = await repl_task.join()
                    response = (result, None)
                except SystemExit:
                    raise
                except BaseException as e:
                    await repl_task.wait()
                    response = (None, e.__cause__)
                await watch_task.cancel()
            finally:
                signal.signal(signal.SIGINT, hand)
            await self.response.put(response)

        self.requests.put(run_it())
        # Get the result here...
        result, exc = self.response.get()
        if exc is not None:
            try:
                raise exc
            except BaseException:
                self.showtraceback()
        else:
            return result

    # Task that runs in the main thread, executing input fed to it from above
    async def runmain(self):
        try:
            hand = signal.signal(signal.SIGINT, signal.SIG_IGN)
            while True:
                coro = await self.requests.get()
                if coro is None:
                    break
                await coro
        finally:
            signal.signal(signal.SIGINT, hand)

def run_repl(console):
    try:
        banner = (
            f'curio REPL {sys.version} on {sys.platform}\n'
            f'Use "await" directly instead of "curio.run()".\n'
            f'Type "help", "copyright", "credits" or "license" '
            f'for more information.\n'
            f'{getattr(sys, "ps1", ">>> ")}import curio'
            )
        console.interact(
            banner=banner,
            exitmsg='exiting curio REPL...')
    finally:
        warnings.filterwarnings(
            'ignore',
            message=r'^coroutine .* was never awaited$',
            category=RuntimeWarning)
        console.requests.put(None)
    
if __name__ == '__main__':
    repl_locals = { 'curio': curio,
                    'ps': curio.monitor.ps,
                    'where': curio.monitor.where,
    }
    for key in {'__name__', '__package__',
                '__loader__', '__spec__',
                '__builtins__', '__file__'}:
        repl_locals[key] = locals()[key]

    console = CurioIOInteractiveConsole(repl_locals)
    threading.Thread(target=run_repl, args=[console], daemon=True).start()
    curio.run(console.runmain)



# curio/task.py
#
# Task class and task related functions such as spawning and cancellation.

__all__ = [
    'Task', 'TaskGroup', 'current_task', 'spawn', 
    'disable_cancellation', 'check_cancellation', 'set_cancellation'
]

# -- Standard library

import logging
from collections import deque
import linecache
import traceback
import os.path

log = logging.getLogger(__name__)

# -- Curio

from .errors import *
from .traps import *
from .sched import SchedBarrier
from . import meta

# Internal functions used for debugging/diagnostics
def _get_stack(coro):
    '''
    Extracts a list of stack frames from a chain of generator/coroutine calls
    '''
    frames = []
    while coro:
        if hasattr(coro, 'cr_frame'):
            f = coro.cr_frame
            coro = coro.cr_await
        elif hasattr(coro, 'ag_frame'):
            f = coro.ag_frame
            coro = coro.ag_await
        elif hasattr(coro, 'gi_frame'):
            f = coro.gi_frame
            coro = coro.gi_yieldfrom
        else:
            # Note: Can't proceed further.  Need the ags_gen or agt_gen attribute
            # from an asynchronous generator.  See https://bugs.python.org/issue32810
            f = None
            coro = None

        if f is not None:
            frames.append(f)
    return frames

# Create a stack traceback for a task
def _format_stack(task, complete=False):
    '''
    Formats a traceback from a stack of coroutines/generators
    '''
    dirname = os.path.dirname(__file__)
    extracted_list = []
    checked = set()
    for f in _get_stack(task.coro):
        lineno = f.f_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        if not complete and os.path.dirname(filename) == dirname:
            continue
        if filename not in checked:
            checked.add(filename)
            linecache.checkcache(filename)
        line = linecache.getline(filename, lineno, f.f_globals)
        extracted_list.append((filename, lineno, name, line))
    if not extracted_list:
        resp = f'No stack for {task!r}'
    else:
        resp = f'Stack for {task!r} (most recent call last):\n'
        resp += ''.join(traceback.format_list(extracted_list))
    return resp

# Return the (filename, lineno) where a task is currently executing
def _where(task):
    dirname = os.path.dirname(__file__)
    for f in _get_stack(task.coro):
        lineno = f.f_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        if os.path.dirname(filename) == dirname:
            continue
        return filename, lineno
    return None, None

class Task(object):
    '''
    The Task class wraps a coroutine and provides some additional attributes
    related to execution state and debugging.  Tasks are not normally 
    instantiated directly. Instead, use spawn().
    '''
    _lastid = 1
    def __init__(self, coro, parent=None):
        # Informational attributes about the task itself
        self.id = Task._lastid
        Task._lastid += 1
        self.parentid = None if parent is None else parent.id  # Parent task id (if any)
        self.coro = coro              # Underlying generator/coroutine
        self.name = getattr(coro, '__qualname__', str(coro))
        self.daemon = False           # Daemonic flag

        # Attributes updated during execution (safe to inspect)
        self.cycles = 0               # Execution cycles completed
        self.state = 'INITIAL'        # Execution state
        self.cancel_func = None       # Cancellation function
        self.future = None            # Pending Future (if any)
        self.sleep = None             # Pending sleep (if any)
        self.timeout = None           # Pending timeout (if any)
        self.joining = SchedBarrier() # Set of tasks waiting to join with this one
        self.cancelled = None         # Has the task been cancelled?
        self.terminated = False       # Has the task actually Terminated?
        self.cancel_pending = None    # Deferred cancellation exception pending (if any)
        self.allow_cancel = True      # Can cancellation exceptions be delivered?
        self.taskgroup = None         # Containing task group (if any)
        self.joined = False           # Set if the task has actually been joined or result collected

        # Final result of coroutine execution (use properties to access)
        self._final_result = None     # Final result of execution
        self._final_exc = None        # Final exception of execution

        # Actual execution is wrapped by a supporting coroutine
        self._run_coro = self._task_runner(self.coro)

        # Result of the last trap
        self._trap_result = None 

        # Last I/O operation performed
        self._last_io = None          

        # Bound coroutine methods
        self._send = self._run_coro.send   
        self._throw = self._run_coro.throw
        
        # Timeout deadline stack
        self._deadlines = []

    def __repr__(self):
        return f'{type(self).__name__}(id={self.id}, name={self.name!r}, state={self.state!r})'

    def __str__(self):
        filename, lineno = _where(self)
        if filename:
            return f'{self!r} at {filename}:{lineno}'
        else:
            return repr(self)

    def __del__(self):
        self.coro.close()
        if not self.joined and not self.cancelled and not self.daemon:
            if not self.daemon and not self.exception:
                log.warning('%r never joined', self)
        assert not self._last_io

    def send(self, value):
        '''
        Send the next value into the task coroutine.  This method is
        a wrapper around the actual method and may be subclassed to
        implement different kinds of low-level functionality.
        '''
        return self._send(value)

    async def _task_runner(self, coro):
        try:
            return await coro
        finally:
            if self.taskgroup:
                self.joined = True
                await self.taskgroup._task_done(self)

    async def join(self):
        '''
        Wait for a task to terminate.  Returns the return value (if any)
        or raises a TaskError if the task crashed with an exception.
        '''
        await self.wait()
        if self.taskgroup:
            self.taskgroup._task_discard(self)
        self.joined = True
        if self.exception:
            raise TaskError('Task crash') from self.exception
        else:
            return self.result

    async def wait(self):
        '''
        Wait for a task to terminate. Does not return any value.
        '''
        if not self.terminated:
            await self.joining.suspend('TASK_JOIN')
        
    @property
    def result(self):
        '''
        Return the result of a task. The task must be terminated already.
        '''
        if not self.terminated:
            raise RuntimeError('Task not terminated')
        self.joined = True
        if self._final_exc:
            raise self._final_exc
        else:
            return self._final_result

    @result.setter
    def result(self, value):
        self._final_result = value
        self._final_exc = None

    @property
    def exception(self):
        '''
        Return any pending exception of a task or None.
        '''
        if not self.terminated:
            raise RuntimeError('Task not terminated')
        return self._final_exc

    @exception.setter
    def exception(self, value):
        self._final_result = None
        self._final_exc = value

    async def cancel(self, *, exc=TaskCancelled, blocking=True):
        '''
        Cancel a task by raising a CancelledError exception.

        If blocking=False, schedules the cancellation and returns immediately.

        If blocking=True (the default), then does not
        return until the task actually terminates.
        '''
        if self.terminated:
            self.joined = True
            return
        await _cancel_task(self, exc=exc)
        if blocking:
            await self.wait()

    def traceback(self):
        '''
        Return a formatted traceback showing where the task is currently executing.
        '''
        return _format_stack(self)

    def where(self):
        '''
        Return a tuple (filename, lineno) where task is executing
        '''
        return _where(self)

class ContextTask(Task):
    '''
    Task class that provides support for contextvars.  Use with the
    taskcls keyword argument to the Curio kernel.
    '''
    def __init__(self, coro, parent=None):
        import contextvars
        super().__init__(coro)
        if parent:
            parent._context.run(lambda: setattr(self, '_context', contextvars.copy_context()))
        else:
            self._context = contextvars.copy_context()

    def send(self, value):
        return self._context.run(super().send, value)

class TaskGroup(object):
    '''
    A TaskGroup represents a collection of managed tasks.  A group can
    be used to ensure that all tasks terminate together, to monitor
    tasks as they finish, and to collect results.

    A TaskGroup can be created from existing tasks.  For example:

        t1 = await spawn(coro1)
        t2 = await spawn(coro2)
        t3 = await spawn(coro3)

        async with TaskGroup([t1,t2,t3]) as g:
            ...

    Alternatively, tasks can be spawned into a task group.

        async with TaskGroup(wait) as g:
            await g.spawn(coro1)
            await g.spawn(coro2)
            await g.spawn(coro3)

    When used as a context manager, a TaskGroup will wait until
    all contained tasks exit before moving on.   The optional wait argument
    specifies a strategy.  If wait=all (the default), a task group
    waits for all tasks to exit.  If wait=any, the group waits
    for the first task to exit. If wait=object, the group waits for 
    the first task to return a non-None result.  If wait=None, the
    group immediately cancels all running tasks.

    Task groups are often used to gather results. The following
    properties are useful:

        async with TaskGroup() as g:
            ...

        print(g.result)    # Result of the first task to exit
        print(g.results)   # List of all results computed

    Note: Both of these properties may raise an exception if a task
    exited in error. The g.tasks property is a list of all
    managed tasks listed in order of creation. 

    To obtain tasks in the order that they complete as they complete,
    use iteration:
  
        async with TaskGroup() as g:
            await g.spawn(coro1)
            await g.spawn(coro2)
            await g.spawn(coro3)

            async for done in g:
                print('Task done', done, done.result)

    The cancel_remaining() method can be used to cancel all
    remaining tasks early.  The add_task() method can be
    used to add an already existing task to a group.  Calling
    .join() on a specific task removes it from a group. For example:

         async with TaskGroup() as g:
             t1 = await g.spawn(coro1)
             ...
             await t1.join()        # removes t1 from the group

    Normally, a task group is used as a context manager.  This 
    doesn't have to be the case.  You could write code like this:

        g = TaskGroup()
        try:
            await g.spawn(coro1)
            await g.spawn(coro2)
            ...
        finally:
            await g.join()
 
    This might be more useful for more persistent or long-lived 
    task groups.

    If any managed task exits with an error, all remaining tasks 
    are cancelled.  The handling of task-related errors takes place when
    the resultof a task group is analyzed.  For example:

        async with TaskGroup(wait=any) as g:
            await g.spawn(coro1)
            await g.spawn(coro2)
            ...
        try:
            result = g.result      # Errors reported here
        except Exception as e:
            print("FAILED:", e)
    '''
    def __init__(self, tasks=(), *, wait=all):
        self._running = set()        # All running tasks
        self._finished = deque()     # All finished tasks
        self._daemonic = set()       # All running daemon tasks
        self._tasks = set()          # Set of all tasks tracked by the group
        self._joined = False
        self._wait = wait            # Wait policy 
        self.completed = None        # First completed task
        for task in tasks:
            assert not task.taskgroup,  "Task already assigned to a task group"
            task.taskgroup = self
            if not task.daemon:
                self._tasks.add(task)
            if task.terminated:
                self._finished.append(task)
            elif task.daemon:
                self._daemonic.add(task)
            else:
                self._running.add(task)

        self._sema = sync.Semaphore(len(self._finished))

    # Property that returns all tracked tasks in task creation order
    @property
    def tasks(self):
        return sorted(self._tasks, key=lambda task: task.id)

    # Property that returns the result of the first completed task
    @property
    def result(self):
        if not self._joined:
            raise RuntimeError("Task group not yet terminated")
        if not self.completed:
            raise RuntimeError("No task successfully completed")
        return self.completed.result

    # Property that returns the exception of the first completed task
    @property
    def exception(self):
        if not self._joined:
            raise RuntimeError("Task group not yet terminated")
        return self.completed.exception if self.completed else None

    # Property that returns all task results (in task creation order)
    @property
    def results(self):
        if not self._joined:
            raise RuntimeError("Task group not yet terminated")
        return [ task.result for task in self.tasks ]

    @property
    def exceptions(self):
        if not self._joined:
            raise RuntimeError("Task group not yet terminated")
        return [ task.exception for task in self.tasks ]
        
    # Triggered on task completion. 
    async def _task_done(self, task):
        if task.daemon:
            self._daemonic.discard(task)
        else:
            self._running.discard(task)
            self._finished.append(task)
            await self._sema.release()

    # Discards a task from the TaskGroup.  Called implicitly if
    # if a task is explicitly joined while under supervision.
    def _task_discard(self, task):
        try:
            self._finished.remove(task)
        except ValueError:
            pass
        self._tasks.discard(task)
        task.taskgroup = None

    async def add_task(self, task):
        '''
        Add an already existing task to the group.
        '''
        if task.taskgroup:
            raise RuntimeError('Task is already part of a task group')

        if self._joined:
            raise RuntimeError("TaskGroup already joined")

        task.taskgroup = self
        if not task.daemon:
            self._tasks.add(task)
        if task.terminated:
            await self._task_done(task)
        elif task.daemon:
            self._daemonic.add(task)
        else:
            self._running.add(task)

    async def spawn(self, coro, *args, daemon=False):
        '''
        Spawn a new task into the task group.
        '''
        if self._joined:
            raise RuntimeError("TaskGroup already joined")
        task = await spawn(coro, *args, daemon=daemon)
        await self.add_task(task)
        return task

    async def spawn_thread(self, func, *args, daemon=False):
        '''
        Spawn a new async thread into a task group
        '''
        from . import thread
        if self._joined:
            raise RuntimeError("TaskGroup already joined")
        thr = await thread.spawn_thread(func, *args, daemon=daemon)
        await self.add_task(thr)
        return thr

    async def next_done(self):
        '''
        Wait for the next task to finish and return it.  This removes it
        from the group.
        '''
        while not self._finished and self._running:
            await self._sema.acquire()

        if self._finished:
            task = self._finished.popleft()
            await task.wait()
            task.taskgroup = None
            self._tasks.discard(task)
        else:
            task = None

        return task

    async def next_result(self):
        '''
        Return the result of the next task that finishes. Note: if task
        terminated via exception, that exception is raised here.
        '''
        task = await self.next_done()
        if task:
            return task.result
        else:
            raise RuntimeError('No task available')

    async def cancel_remaining(self):
        '''
        Cancel all remaining non-daemonic tasks. Tasks are removed
        from the task group when explicitly cancelled.
        '''
        running = list(self._running)
        for task in running:
            await task.cancel(blocking=False)
        for task in running:
            await task.wait()
            self._task_discard(task)

            
    async def join(self):
        '''
        Wait for tasks in a task group to terminate according to the
        wait policy set for the group.  
        '''
        try:
            if self._wait is None:
                # We wait for no-one. Tasks get cancelled on return.
                return

            while self._finished or self._running:
                # If nothing is finished, we wait for something to complete
                while not self._finished:
                    await self._sema.acquire()

                # Examine all currently finished tasks
                while self._finished:
                    task = self._finished.popleft()
                    # Check if it's the first completed task
                    if self.completed is None:
                        # For wait=object, the self.completed attribute is the first non-None result
                        if not ((self._wait is object) and (not task.exception) and (task.result is None)):
                            self.completed = task 

                    # What happens next depends on the wait and error handling policies
                    if task.exception or \
                       (self._wait is any) or \
                       ((self._wait is object) and (task.result is not None)):
                        return

        # Task groups guarantee all tasks cancelled/terminated upon join()
        finally:
            while self._running:
                task = self._running.pop()
                await task.cancel()
            while self._daemonic:
                task = self._daemonic.pop()
                await task.cancel()
            self._joined = True
        return

    async def __aenter__(self):
        return self

    async def __aexit__(self, ty, val, tb):
        if ty:
            await self.cancel_remaining()
        await self.join()

    def __aiter__(self):
        return self

    async def __anext__(self):
        next = await self.next_done()
        if next is None:
            raise StopAsyncIteration
        return next

    # -- Support for use in async threads
    def __enter__(self):
        return thread.AWAIT(self.__aenter__())

    def __exit__(self, *args):
        return thread.AWAIT(self.__aexit__(*args))

    def __iter__(self):
        return thread.AWAIT(self.__aiter__())

    def __next__(self):
        try:
            return thread.AWAIT(self.__anext__())
        except StopAsyncIteration:
            raise StopIteration

# ----------------------------------------------------------------------
# Public-facing task-related functions.  Some of these functions are
# merely a layer over a low-level trap using async/await.  One reason
# for doing this is that the user will get a more proper warning message
# if they use the function without using the required 'await' keyword.
# -----------------------------------------------------------------------

async def current_task():
    '''
    Returns a reference to the current task
    '''
    return await _get_current()

async def spawn(corofunc, *args, daemon=False):
    '''
    Create a new task, running corofunc(*args). Use the daemon=True
    option if the task runs forever as a background task. 
    '''
    coro = meta.instantiate_coroutine(corofunc, *args)
    task = await _spawn(coro)
    task.daemon = daemon
    return task

# Context manager for supervising cancellation masking
class _CancellationManager(object):

    async def __aenter__(self):
        self.task = await current_task()
        self._last_allow_cancel = self.task.allow_cancel
        self.task.allow_cancel = False
        return self

    async def __aexit__(self, ty, val, tb):
        # Restore previous cancellation flag
        self.task.allow_cancel = self._last_allow_cancel

        # If a CancelledError is being raised on exit from a block, it
        # becomes pending in the task if cancellation is not allowed
        # in the outer context.  Curio should never have delivered 
        # such an exception on its own--it could be manually raised however.
        if isinstance(val, CancelledError) and not self.task.allow_cancel:
            self.task.cancel_pending = val
            return True
        else:
            return False

    def __enter__(self):
        return thread.AWAIT(self.__aenter__())

    def __exit__(self, *args):
        return thread.AWAIT(self.__aexit__(*args))

def disable_cancellation(coro=None, *args):
    if coro is None:
        return _CancellationManager()
    else:
        coro = meta.instantiate_coroutine(coro, *args)
        async def run():
            async with _CancellationManager():
                return await coro
        return run()

async def check_cancellation(exc_type=None):
    '''
    Check if there is any kind of pending cancellation. If cancellations
    are currently allowed, and there is a pending exception, it raises the
    exception.  If cancellations are not allowed, it returns the pending
    exception object or None..

    If exc_type is specified, the function checks the type of the specified
    exception against the given type.  If there is a match, the exception
    is returned and cleared.
    '''
    task = await current_task()

    if exc_type and not isinstance(task.cancel_pending, exc_type):
        return None

    if task.cancel_pending and task.allow_cancel:
        try:
            raise task.cancel_pending
        finally:
            task.cancel_pending = None
    else:
        try:
            return task.cancel_pending
        finally:
            if exc_type:
                task.cancel_pending = None

async def set_cancellation(exc):
    '''
    Set a new pending cancellation exception. Returns the old exception.
    '''
    task = await current_task()
    result = task.cancel_pending
    task.cancel_pending = exc
    return result

# Here to avoid circular import issues
from . import thread
from . import sync





# curio/workers.py
#
# Functions for performing work outside of curio.  This includes
# running functions in threads, processes, and executors from the
# concurrent.futures module.

__all__ = ['run_in_executor', 'run_in_thread', 'run_in_process', 'block_in_thread']

# -- Standard Library

import sys
import multiprocessing
import threading
import traceback
import signal
from collections import Counter, defaultdict

# -- Curio

from .errors import CancelledError
from .traps import _future_wait, _get_kernel
from . import sync
from .channel import Connection

# Code to embed a traceback in a remote exception.  This is borrowed
# straight from multiprocessing.pool.  Copied here to avoid possible
# confusion when reading the traceback message (it will identify itself
# as originating from curio as opposed to multiprocessing.pool).

class RemoteTraceback(Exception):

    def __init__(self, tb):
        self.tb = tb

    def __str__(self):
        return self.tb


class ExceptionWithTraceback:

    def __init__(self, exc, tb):
        tb = traceback.format_exception(type(exc), exc, tb)
        tb = ''.join(tb)
        self.exc = exc
        self.tb = '\n"""\n%s"""' % tb

    def __reduce__(self):
        return rebuild_exc, (self.exc, self.tb)


def rebuild_exc(exc, tb):
    exc.__cause__ = RemoteTraceback(tb)
    return exc

async def run_in_executor(exc, callable, *args):
    '''
    Run callable(*args) in an executor such as
    ThreadPoolExecutor or ProcessPoolExecutor from the
    concurrent.futures module.  Be aware that on cancellation, any
    worker thread or process that was handling the request will
    continue to run to completion as a kind of zombie-- possibly
    rendering the executor unusable for subsequent work.

    This function is provided for compatibility with
    concurrent.futures, but is not the recommend approach for running
    blocking or cpu-bound work in curio. Use the run_in_thread() or
    run_in_process() methods instead.
    '''
    future = exc.submit(callable, *args)
    await _future_wait(future)
    return future.result()

MAX_WORKER_THREADS = 64

async def reserve_thread_worker():
    '''
    Reserve a thread pool worker
    '''
    kernel = await _get_kernel()
    if not hasattr(kernel, 'thread_pool'):
        kernel.thread_pool = WorkerPool(ThreadWorker, MAX_WORKER_THREADS)
        kernel._call_at_shutdown(kernel.thread_pool.shutdown)
    return (await kernel.thread_pool.reserve())

async def run_in_thread(callable, *args, call_on_cancel=None):
    '''
    Run callable(*args) in a separate thread and return the result. If
    cancelled, be aware that the requested callable may or may not have
    executed.  If it start running, it will run fully to completion
    as a kind of zombie.
    '''
    worker = None
    try:
        worker = await reserve_thread_worker()
        return await worker.apply(callable, args, call_on_cancel)
    finally:
        if worker:
            await worker.release()

# Support for blocking in threads.
#
# Discussion:
#
# The run_in_thread() function can be used to run any synchronous function
# in a separate thread.  However, certain kinds of operations are
# inherently unsafe.   For example, consider a worker task that wants
# to wait on a threading Event like this:
#
#    evt = threading.Event()     # Foreign Event...
#
#    async def worker():
#        await run_in_thread(evt.wait)
#        print('Alive!')
#
# Now suppose Curio spins up a huge number of workers:
#
#    for n in range(1000):
#        await spawn(worker())
#
# At this point, you're in a bad situation.  The worker tasks have all
# called run_in_thread() and are blocked indefinitely.  Because the
# pool of worker threads is limited, you've exhausted all available
# resources.  Nobody can now call run_in_thread() without blocking.
# There's a pretty good chance that your code is permanently
# deadlocked.  There are dark clouds.
#
# This problem can be solved by wrapping run_in_thread() with a
# semaphore. Like this:
#
#    _barrier = curio.Semaphore()
#
#    async def worker():
#        async with _barrier:
#            await run_in_thread(evt.wait)
#
# However, to make it much more convenient, we can take care of
# a lot of fiddly details.  We can cache the requested callable,
# build a set of semaphores and synchronize things in the background.
# That's what the block_in_thread() function is doing.  For example:
#
#    async def worker():
#        await block_in_thread(evt.wait)
#        print('Alive!')
#
# Unlike run_in_thread(), spawning up 1000 workers creates a
# situation where only 1 worker is actually blocked in a thread.
# The other 999 workers are blocked on a semaphore waiting for service.

_pending = Counter()
_barrier = defaultdict(sync.Semaphore)

async def block_in_thread(callable, *args, call_on_cancel=None):
    '''
    Run callable(*args) in a thread with the expectation that the
    operation is going to block for an indeterminate amount of time.
    Guarantees that at most only one background thread is used
    regardless of how many curio tasks are actually waiting on the
    same callable (e.g., if 1000 Curio tasks all decide to call
    block_on_thread on the same callable, they'll all be handled by a
    single thread). Primary use of this function is on foreign locks,
    queues, and other synchronization primitives where you have to use
    a thread, but you just don't have any idea when the operation will
    complete.
    '''
    if hasattr(callable, '__self__'):
        call_key = (callable.__name__, id(callable.__self__))
    else:
        call_key = id(callable)
    _pending[call_key] += 1
    async with _barrier[call_key]:
        try:
            return await run_in_thread(callable, *args, call_on_cancel=call_on_cancel)
        finally:
            _pending[call_key] -= 1
            if not _pending[call_key]:
                del _pending[call_key]
                del _barrier[call_key]


MAX_WORKER_PROCESSES = multiprocessing.cpu_count()

async def run_in_process(callable, *args):
    '''
    Run callable(*args) in a separate process and return the
    result.  In the event of cancellation, the worker process is
    immediately terminated.

    The worker process is created using multiprocessing.Process().
    Communication with the process uses multiprocessing.Pipe() and an
    asynchronous message passing channel.  All function arguments and
    return values are seralized using the pickle module.  When
    cancelled, the Process.terminate() method is used to kill the
    worker process.  This results in a SIGTERM signal being sent to
    the process.

    The handle_cancellation flag, if True, indicates that you intend
    to manage the worker cancellation yourself.  This an advanced
    option.  Any resulting CancelledError has 'task' and 'worker'
    attributes.  task is a background task that's supervising the
    still executing work.  worker is the associated process.

    The worker process is a separate isolated Python interpreter.
    Nothing should be assumed about its global state including shared
    variables, files, or connections.
    '''
    kernel = await _get_kernel()
    if not hasattr(kernel, 'process_pool'):
        kernel.process_pool = WorkerPool(ProcessWorker, MAX_WORKER_PROCESSES)
        kernel._call_at_shutdown(kernel.process_pool.shutdown)
    worker = None
    try:
        worker = await kernel.process_pool.reserve()
        return await worker.apply(callable, args)
    finally:
        if worker:
            await worker.release()

# The _FutureLess class is a custom "Future" implementation solely for
# use by curio. It is used by the ThreadWorker class below and
# provides only the minimal set of functionality needed to transmit a
# result back to the curio kernel.  Unlike the normal Future class,
# this version doesn't require any thread synchronization or
# notification support.  By eliminating that, the overhead associated
# with the handoff between curio tasks and threads is substantially
# faster.


class _FutureLess(object):
    __slots__ = ('_callback', '_exception', '_result')

    def set_result(self, result):
        self._result = result
        self._callback(self)

    def set_exception(self, exc):
        self._exception = exc
        self._callback(self)

    def result(self):
        try:
            return self._result
        except AttributeError:
            raise self._exception from None

    def add_done_callback(self, func):
        self._callback = func

    def cancel(self):
        pass

# A ThreadWorker represents a thread that performs work on behalf of a
# curio task.   A curio task initiates work by executing the
# apply() method. This passes the request to a background thread that
# executes it.  While this takes place, the curio task blocks, waiting
# for a result to be set on an internal Future.


class ThreadWorker(object):
    '''
    Worker that executes a callable on behalf of a curio task in a separate thread.
    '''

    def __init__(self, pool):
        self.thread = None
        self.start_evt = None
        self.lock = None
        self.request = None
        self.terminated = False
        self.pool = pool

    def _launch(self):
        self.start_evt = threading.Event()
        self.thread = threading.Thread(target=self.run_worker, daemon=True)
        self.thread.start()

    def run_worker(self):
        while True:
            self.start_evt.wait()
            self.start_evt.clear()
            # If there is no pending request, but we were signalled to
            # start, it means terminate.
            if not self.request:
                return

            # Run the request
            self.request()

    async def release(self):
        if self.pool:
            await self.pool.release(self)

    def shutdown(self):
        self.terminated = True
        self.request = None
        if self.start_evt:
            self.start_evt.set()

    async def apply(self, func, args=(), call_on_cancel=None):
        '''
        Run the callable func in a separate thread and return the result.
        '''
        if self.thread is None:
            self._launch()

        # Set up a request for the worker thread
        done_evt = threading.Event()
        done_evt.clear()
        cancelled = False
        future = _FutureLess()

        def run_callable():
            try:
                future.set_result(func(*args))
            except BaseException as err:
                future.set_exception(err)
            finally:
                done_evt.wait()
                if cancelled and call_on_cancel:
                    call_on_cancel(future)

        self.request = run_callable
        try:
            await _future_wait(future, self.start_evt)
            return future.result()
        except CancelledError as e:
            cancelled = True
            self.shutdown()
            raise
        finally:
            done_evt.set()

class ProcessWorker(object):
    '''
    Managed process worker for running CPU-intensive tasks.  The main
    purpose of this class is to run workers with reliable
    cancellation/timeout semantics. Specifically, if a worker is
    cancelled, the underlying process is also killed.   This, as
    opposed to having it linger on running until work is complete.
    '''
    def __init__(self, pool):
        self.process = None
        self.client_ch = None
        self.terminated = False
        self.pool = pool

    def _launch(self):
        context = multiprocessing.get_context('spawn')
        client_ch, server_ch = context.Pipe()
        self.process = context.Process(
            target=self.run_server, args=(server_ch, ), daemon=True)
        self.process.start()
        server_ch.close()
        self.client_ch = Connection.from_Connection(client_ch)

    def shutdown(self):
        self.terminated = True
        if self.process:
            self.process.terminate()
            self.process = None
            self.nrequests = 0

    async def release(self):
        if self.pool:
            await self.pool.release(self)

    def run_server(self, ch):
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        while True:
            func, args = ch.recv()
            try:
                result = func(*args)
                ch.send((True, result))
            except Exception as e:
                e = ExceptionWithTraceback(e, e.__traceback__)
                ch.send((False, e))
            del func, args

    async def apply(self, func, args=()):
        if self.process is None or not self.process.is_alive():
            self._launch()

        msg = (func, args)
        try:
            await self.client_ch.send(msg)
            success, result = await self.client_ch.recv()
            if success:
                return result
            else:
                raise result
        except CancelledError:
            self.shutdown()
            raise

# Windows-compatible process worker.  It differs from ProcessWorker in
# that client communication is handled synchronously by a thread.
class WinProcessWorker(ProcessWorker):
    def _launch(self):
        context = multiprocessing.get_context('spawn')
        client_ch, server_ch = context.Pipe()
        self.process = context.Process(
            target=self.run_server, args=(server_ch, ), daemon=True)
        self.process.start()
        server_ch.close()
        self.client_ch = client_ch

    def _client_communicate(self, msg):
        self.client_ch.send(msg)
        return self.client_ch.recv()

    async def apply(self, func, args=()):
        if self.process is None or not self.process.is_alive():
            self._launch()

        msg = (func, args)
        try:
            success, result = await run_in_thread(self._client_communicate, msg)
            if success:
                return result
            else:
                raise result
        except CancelledError:
            self.shutdown()
            raise

if sys.platform.startswith('win'):
    ProcessWorker = WinProcessWorker

# Pool of workers for carrying out jobs on behalf of curio tasks.
#
# This pool works a bit differently than a normal thread/process
# pool due to some of the different ways that threads get used in Curio.
# Instead of submitting work to the pool, you use the reserve() method
# to obtain a worker:
#
#     worker = await pool.reserve()
#
# Once you have a worker, it is yours for as long as you want to have
# it.  To submit work to it, use the apply() method:
#
#     await worker.apply(callable, args)
#
# When you're done with it, release it back to the pool.
#
#     await worker.release()
#
# Some rationale for this design:  Sometimes when you're working with
# threads, you want to perform multiple steps and you need to make sure
# you're performing each step on the same thread for some reason. This
# is especially true if you're trying to manage work cancellation.
# For example, work started in a thread might need to be cleaned up
# on the same thread.  By reserving/releasing workers, we get more
# control over the whole process of how workers get managed.

class WorkerPool(object):

    def __init__(self, workercls, nworkers):
        self.nworkers = sync.Semaphore(nworkers)
        self.workercls = workercls
        self.workers = []

    def shutdown(self):
        for worker in self.workers:
            worker.shutdown()
        self.workers = []

    async def reserve(self):
        await self.nworkers.acquire()
        if not self.workers:
            return self.workercls(self)
        else:
            return self.workers.pop()

    async def release(self, worker):
        if not worker.terminated:
            self.workers.append(worker)
        await self.nworkers.release()


# Pool definitions should anyone want to use them directly
ProcessPool = lambda nworkers: WorkerPool(ProcessWorker, nworkers)
ThreadPool = lambda nworkers: WorkerPool(ThreadWorker, nworkers)



# curio/network.py
#
# Some high-level functions useful for writing network code.  These are loosely
# based on their similar counterparts in the asyncio library. Some of the
# fiddly low-level bits are borrowed.

__all__ = [ 'open_connection', 'tcp_server', 'tcp_server_socket',
            'open_unix_connection', 'unix_server', 'unix_server_socket' ]

# -- Standard library

import logging
log = logging.getLogger(__name__)

# -- Curio

from . import socket
from . import ssl as curiossl
from .task import TaskGroup
from .io import Socket


async def _wrap_ssl_client(sock, ssl, server_hostname, alpn_protocols):
    # Applies SSL to a client connection. Returns an SSL socket.
    if ssl:
        if isinstance(ssl, bool):
            sslcontext = curiossl.create_default_context()
            if not server_hostname:
                sslcontext._context.check_hostname = False
                sslcontext._context.verify_mode = curiossl.CERT_NONE

            if alpn_protocols:
                sslcontext.set_alpn_protocols(alpn_protocols)
        else:
            # Assume that ssl is an already created context
            sslcontext = ssl

        if server_hostname:
            extra_args = {'server_hostname': server_hostname}
        else:
            extra_args = {}

        # if the context is Curio's own, it expects a Curio socket and
        # returns one. If context is from an external source, including
        # the stdlib's ssl.SSLContext, it expects a non-Curio socket and
        # returns a non-Curio socket, which then needs wrapping in a Curio
        # socket.
        #
        # Perhaps the CurioSSLContext is no longer needed. In which case,
        # this code can be simplified to just the else case below.
        #
        if isinstance(sslcontext, curiossl.CurioSSLContext):
            sock = await sslcontext.wrap_socket(sock, do_handshake_on_connect=False, **extra_args)
        else:
            # do_handshake_on_connect should not be specified for
            # non-blocking sockets
            extra_args['do_handshake_on_connect'] = sock._socket.gettimeout() != 0.0
            sock = Socket(sslcontext.wrap_socket(sock._socket, **extra_args))
        await sock.do_handshake()
    return sock

async def open_connection(host, port, *, ssl=None, source_addr=None, server_hostname=None,
                          alpn_protocols=None):
    '''
    Create a TCP connection to a given Internet host and port with optional SSL applied to it.
    '''
    if server_hostname and not ssl:
        raise ValueError('server_hostname is only applicable with SSL')

    sock = await socket.create_connection((host, port), source_address=source_addr)

    try:
        # Apply SSL wrapping to the connection, if applicable
        if ssl:
            sock = await _wrap_ssl_client(sock, ssl, server_hostname, alpn_protocols)

        return sock
    except Exception:
        sock._socket.close()
        raise

async def open_unix_connection(path, *, ssl=None, server_hostname=None,
                               alpn_protocols=None):
    if server_hostname and not ssl:
        raise ValueError('server_hostname is only applicable with SSL')

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        await sock.connect(path)

        # Apply SSL wrapping to connection, if applicable
        if ssl:
            sock = await _wrap_ssl_client(sock, ssl, server_hostname, alpn_protocols)

        return sock
    except Exception:
        sock._socket.close()
        raise

async def run_server(sock, client_connected_task, ssl=None):
    if ssl and not hasattr(ssl, 'wrap_socket'):
        raise ValueError('ssl argument must have a wrap_socket method')

    async def run_client(client, addr):
        async with client:
            await client_connected_task(client, addr)

    async def run_server(sock, group):
        while True:
            client, addr = await sock.accept()
            if ssl:
                if isinstance(ssl, curiossl.CurioSSLContext):
                    client = await ssl.wrap_socket(client, server_side=True, do_handshake_on_connect=False)
                else:
                    client = ssl.wrap_socket(client, server_side=True, do_handshake_on_connect=False)
                if not isinstance(client, Socket):
                    client = Socket(client)
            await group.spawn(run_client, client, addr)
            del client

    async with sock:
        async with TaskGroup() as tg:
            await tg.spawn(run_server, sock, tg)
            # Reap all of the children tasks as they complete
            async for task in tg:
                task.joined = True
                del task

def tcp_server_socket(host, port, family=socket.AF_INET, backlog=100,
                      reuse_address=True, reuse_port=False):

    sock = socket.socket(family, socket.SOCK_STREAM)
    try:
        if reuse_address:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

        if reuse_port:
            try:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, True)
            except (AttributeError, OSError) as e:
                log.warning('reuse_port=True option failed', exc_info=True)

        sock.bind((host, port))
        sock.listen(backlog)
    except Exception:
        sock._socket.close()
        raise

    return sock

async def tcp_server(host, port, client_connected_task, *,
                     family=socket.AF_INET, backlog=100, ssl=None,
                     reuse_address=True, reuse_port=False):

    sock = tcp_server_socket(host, port, family, backlog, reuse_address, reuse_port)
    await run_server(sock, client_connected_task, ssl)

def unix_server_socket(path, backlog=100):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        sock.bind(path)
        sock.listen(backlog)
    except Exception:
        sock._socket.close()
        raise
    return sock

async def unix_server(path, client_connected_task, *, backlog=100, ssl=None):
    sock = unix_server_socket(path, backlog)
    await run_server(sock, client_connected_task, ssl)



# curio/kernel.py
#
# Main execution kernel.
#
# Curio is based on a few overarching design principles that drive the code
# you'll find here.
#
# 1. Environmental Isolation.
#
#    Curio strictly separates the environment of async and synchronous
#    programming.  All functionality related to async operation is
#    placed in async-function definitions.  Async functions request
#    the services of the kernel using low-level yield statements
#    (traps).  The kernel is an opaque black-box from the perspective
#    of synchronous code.  There is only one available
#    operation--run(coro) which runs a new task.  There are no other
#    mechanisms available for interacting with the kernel from
#    synchronous code.  A good analogy might be the distinction
#    between user and protected mode in an OS.  User programs run in
#    user-mode and the operating system kernel runs in protected mode.
#    The same thing happens here.  User programs in Curio can only run
#    in async functions. Those programs can request the services of
#    the kernel. However, they're not granted any further access than
#    that (there is no API surface or anything that can be used).
#
# 2. Microkernels
#
#    The low-level kernel is meant to be small, fast, and minimally
#    featureful.  In fact, almost nothing interesting happens in the
#    kernel. Instead, almost every useful part of Curio gets
#    implemented in async functions found elsewhere.  If you're trying
#    to add new features to Curio, don't add them to the kernel. Think
#    about how to create objects and functions that operate at the
#    async-function level instead.  See files such as sync.py or
#    queue.py for examples.
#
# 3. Decoupling
#
#    No part of Curio has direct linkage to the Kernel class (it's
#    not imported or used anywhere else in the code base).   If you want,
#    you can make a completely custom Kernel object and have the
#    rest of Curio run on it.  You just need to make sure you implement
#    the required traps.   This is in contrast to libraries such as
#    asyncio where many parts of the implementation are required to
#    carry a reference to the underlying event loop.

__all__ = [ 'Kernel', 'run' ]

# -- Standard Library

import socket
import time
import os
import errno
from selectors import DefaultSelector, EVENT_READ, EVENT_WRITE
from collections import deque
import threading

# Logger where uncaught exceptions from crashed tasks are logged
import logging
log = logging.getLogger(__name__)

# -- Curio

from .errors import *
from .task import Task
from .traps import _read_wait
from . import meta
from .timequeue import TimeQueue


class Kernel(object):
    '''
    Curio run-time kernel.  The selector argument specifies a
    different I/O selector. The debug argument specifies a list of
    debugger objects to apply. For example:

        from curio.debug import schedtrace, traptrace
        k = Kernel(debug=[schedtrace, traptrace])

    Use the kernel run() method to submit work to the kernel.
    '''

    def __init__(self, *, selector=None, debug=None, activations=None, taskcls=Task,
                 max_select_timeout=None if os.name != 'nt' else 1.0):

        # Functions to call at shutdown
        self._shutdown_funcs = []

        # I/O Selector setup
        self._selector = selector if selector else DefaultSelector()
        self._call_at_shutdown(self._selector.close)

        # Task table
        self._tasks = {}

        # Coroutine runner function (created upon first call to run())
        self._runner = None

        # Activations
        self._activations = activations if activations else []

        # Debugging (activations in disguise)
        if debug:
            from .debug import _create_debuggers
            self._activations.extend(_create_debuggers(debug))

        # Task creation class
        self._taskcls = taskcls

        self._max_select_timeout = max_select_timeout


    def __del__(self):
        if self._shutdown_funcs is not None:
            raise RuntimeError(
                'Curio kernel not properly terminated.  Please use Kernel.run(shutdown=True)')

    def __enter__(self):
        return self

    def __exit__(self, ty, val, tb):
        if self._shutdown_funcs is not None:
            self.run(shutdown=True)

    def _call_at_shutdown(self, func):
        self._shutdown_funcs.append(func)


    # ----------
    # Submit a new task to the kernel

    def run(self, corofunc=None, *args, shutdown=False):
        if self._shutdown_funcs is None:
            raise RuntimeError("Can't run a kernel that's been shut down or crashed. Create a new kernel.")

        coro = meta.instantiate_coroutine(corofunc, *args) if corofunc else None
        with meta.running(self):
            # Make the kernel runtime environment (if needed)
            if not self._runner:
                self._runner = self._make_kernel_runtime()

            ret_val = ret_exc = None
            # Run the supplied coroutine (if any)
            if coro or not shutdown:
                task = self._runner(coro)
                if task:
                    ret_exc = task.exception
                    ret_val = task.result if not ret_exc else None
                del task

            # If shutdown has been requested, run the shutdown process
            if shutdown:
                # For "reasons" related to task scheduling, the task
                # of shutting down all remaining tasks is best managed
                # by a launching a task dedicated to carrying out the task (sic)
                async def _shutdown_tasks(tocancel):
                    for task in tocancel:
                        await task.cancel()

                tocancel = sorted(self._tasks.values(), key=lambda t: t.id, reverse=True)
                self._runner(_shutdown_tasks(tocancel))
                assert not self._tasks, "New tasks created during shutdown"
                self._runner = None

                # Call registered shutdown functions
                for func in self._shutdown_funcs:
                    func()
                self._shutdown_funcs = None

            if ret_exc:
                raise ret_exc
            else:
                return ret_val

    # ------------------------------------------------------------
    # Kernel runtime
    #
    # This function creates the kernel execution environment. It
    # returns a single function (a closure) that executes a coroutine.
    #
    # At first glance, this function is going to look giant and
    # insane. It is implementing the kernel runtime as a self-contained
    # black box.  There is no external API.  The only possible
    # communication is via traps defined in curio/traps.py.
    # It's best to think of this as a "program within a program".

    def _make_kernel_runtime(kernel):

        # Motto:  "What happens in the kernel stays in the kernel"

        # ---- Kernel State
        current = None                          # Currently running task
        selector = kernel._selector             # Event selector
        ready = deque()                         # Ready queue
        tasks = kernel._tasks                   # Task table
        sleepq = TimeQueue()                    # Sleeping task queue
        wake_queue = deque()                    # Thread wake queue
        _activations = []

        # ---- Bound methods
        selector_register = selector.register
        selector_unregister = selector.unregister
        selector_modify = selector.modify
        selector_select = selector.select
        selector_getkey = selector.get_key
        selector_max_timeout = kernel._max_select_timeout

        ready_popleft = ready.popleft
        ready_append = ready.append
        time_monotonic = time.monotonic
        taskcls = kernel._taskcls

        # ------------------------------------------------------------
        # In-kernel task used for processing futures.
        #
        # Internal task that monitors the loopback socket--allowing the kernel to
        # awake for non-I/O events.

        # Loop-back sockets
        notify_sock = None
        wait_sock = None

        async def _kernel_task():
            wake_queue_popleft = wake_queue.popleft
            while True:
                await _read_wait(wait_sock)
                data = wait_sock.recv(1000)

                # Process any waking tasks.  These are tasks that have
                # been awakened externally to the event loop (e.g., by
                # separate threads, Futures, etc.).
                while wake_queue:
                    task, future = wake_queue_popleft()
                    # If the future associated with wakeup no longer
                    # matches the future stored on the task, wakeup is
                    # abandoned.  It means that a timeout or
                    # cancellation event occurred in the time interval
                    # between the call to wake() and the
                    # subsequent processing of the waking task
                    if future and task.future is not future:
                        continue
                    task.future = None
                    task.state = 'READY'
                    task.cancel_func = None
                    ready_append(task)

        # Force the kernel to wake, possibly scheduling a task to run.
        # This method is called by threads running concurrently to the
        # curio kernel.  For example, it's triggered upon completion of
        # Futures created by thread pools and processes. It's inherently
        # dangerous for any kind of operation on the kernel to be
        # performed by a separate thread.  Thus, the *only* thing that
        # happens here is that the task gets appended to a deque and a
        # notification message is written to the kernel notification
        # socket.  append() and pop() operations on deques are thread safe
        # and do not need additional locking.  See
        # https://docs.python.org/3/library/collections.html#collections.deque
        # ----------
        def wake(task=None, future=None):
            if task:
                wake_queue.append((task, future))

            notify_sock.send(b'\x00')

        def init_loopback():
            nonlocal notify_sock, wait_sock
            notify_sock, wait_sock = socket.socketpair()
            wait_sock.setblocking(False)
            notify_sock.setblocking(False)
            kernel._call_at_shutdown(notify_sock.close)
            kernel._call_at_shutdown(wait_sock.close)

        # ------------------------------------------------------------
        # Task management functions.
        #

        # Create a new task. Putting it on the ready queue
        def new_task(coro):
            task = taskcls(coro, current)
            tasks[task.id] = task
            reschedule_task(task)
            for a in _activations:
                a.created(task)
            return task

        # Reschedule a task, putting it back on the ready queue.
        def reschedule_task(task):
            assert task not in ready

            ready_append(task)
            task.state = 'READY'
            task.cancel_func = None

        # Suspend the current task
        def suspend_task(state, cancel_func):
            nonlocal current
            current.state = state
            current.cancel_func = cancel_func

            # Unregister previous I/O request. Discussion follows:
            #
            # When a task performs I/O, it registers itself with the underlying
            # I/O selector.  When the task is reawakened, it unregisters itself
            # and prepares to run.  However, in many network applications, the
            # task will perform a small amount of work and then go to sleep on
            # exactly the same I/O resource that it was waiting on before. For
            # example, a client handling task in a server will often spend most
            # of its time waiting for incoming data on a single socket.
            #
            # Instead of always unregistering the task from the selector, we
            # can defer the unregistration process until after the task goes
            # back to sleep again.  If it happens to be sleeping on the same
            # resource as before, there's no need to unregister it--it will
            # still be registered from the last I/O operation.
            #
            # The code here performs the unregister step for a task that
            # ran, but is now sleeping for a *different* reason than repeating the
            # prior I/O operation.  There is coordination with code in trap_io().

            if current._last_io:
                unregister_event(*current._last_io)
                current._last_io = None

            current = None

        # Check if task has pending cancellation
        def check_cancellation():
            if current.allow_cancel and current.cancel_pending:
                current._trap_result = current.cancel_pending
                current.cancel_pending = None
                return True
            else:
                return False

        # Set a timeout or sleep event on the current task
        def set_timeout(clock, sleep_type='timeout'):
            if clock is None:
                sleepq.cancel((current.id, sleep_type), getattr(current, sleep_type))
            else:
                sleepq.push((current.id, sleep_type), clock)
            setattr(current, sleep_type, clock)

        # ------------------------------------------------------------
        # I/O Support functions
        #

        def register_event(fileobj, event, task):
            try:
                key = selector_getkey(fileobj)
                mask, (rtask, wtask) = key.events, key.data
                if event == EVENT_READ and rtask:
                    raise ReadResourceBusy(f"Multiple tasks can't wait to read on the same file descriptor {fileobj}")
                if event == EVENT_WRITE and wtask:
                    raise WriteResourceBusy(f"Multiple tasks can't wait to write on the same file descriptor {fileobj}")

                selector_modify(fileobj, mask | event,
                                (task, wtask) if event == EVENT_READ else (rtask, task))
                selector_getkey(fileobj)
            except KeyError:
                selector_register(fileobj, event,
                                  (task, None) if event == EVENT_READ else (None, task))

        def unregister_event(fileobj, event):
            key = selector_getkey(fileobj)
            mask, (rtask, wtask) = key.events, key.data
            mask &= ~event
            if not mask:
                selector_unregister(fileobj)
            else:
                selector_modify(fileobj, mask,
                                (None, wtask) if event == EVENT_READ else (rtask, None))

        # ------------------------------------------------------------
        # Traps
        #
        # These implement the low-level functionality that is
        # triggered by user-level code.  They are never invoked directly
        # and there is no public API outside the kernel.  Instead,
        # coroutines use a statement such as
        #
        #   yield ('trap_io', sock, EVENT_READ, 'READ_WAIT')
        #
        # to invoke a specific trap.
        # ------------------------------------------------------------

        # ----------------------------------------
        # Wait for I/O
        def trap_io(fileobj, event, state):
            if check_cancellation():
                return

            # See comment about deferred unregister in suspend_task(). If the
            # requested I/O operation is *different* than the last I/O operation
            # that was performed by the task, we need to unregister the last I/O
            # resource used and register a new one with the selector.
            if current._last_io != (fileobj, event):
                if current._last_io:
                    unregister_event(*current._last_io)
                try:
                    register_event(fileobj, event, current)
                except CurioError as e:
                    current._trap_result = e
                    return

            # This step indicates that we have managed any deferred I/O management
            # for the task.  Otherwise, I/O will be unregistered.
            current._last_io = None
            suspend_task(state, lambda: unregister_event(fileobj, event))

        # ----------------------------------------
        # Release any kernel resources associated with fileobj.
        def trap_io_release(fileobj):
            if current._last_io:
                unregister_event(*current._last_io)
                current._last_io = None
            current._trap_result = None

        # ----------------------------------------
        # Return tasks currently waiting on a file obj.
        def trap_io_waiting(fileobj):
            try:
                key = selector_getkey(fileobj)
                rtask, wtask = key.data
                rtask = rtask if rtask and rtask.cancel_func else None
                wtask = wtask if wtask and wtask.cancel_func else None
                current._trap_result = (rtask, wtask)
            except KeyError:
                current._trap_result = (None, None)

        # ----------------------------------------
        # Wait on a Future
        def trap_future_wait(future, event):
            if check_cancellation():
                return

            current.future = future

            # Discussion: Each task records the future that it is
            # currently waiting on.  The completion callback below only
            # attempts to wake the task if its stored Future is exactly
            # the same one that was stored above.  Due to support for
            # cancellation and timeouts, it's possible that a task might
            # abandon its attempt to wait for a Future and go on to
            # perform other operations, including waiting for different
            # Future in the future (got it?).  However, a running thread
            # or process still might go on to eventually complete the
            # earlier work.  In that case, it will trigger the callback,
            # find that the task's current Future is now different, and
            # discard the result.

            future.add_done_callback(lambda fut, task=current: wake(task, fut))

            # An optional threading.Event object can be passed and set to
            # start a worker thread.   This makes it possible to have a lock-free
            # Future implementation where worker threads only start after the
            # callback function has been set above.
            if event:
                event.set()

            suspend_task('FUTURE_WAIT',
                          lambda task=current:
                              setattr(task, 'future', future.cancel() and None))

        # ----------------------------------------
        # Add a new task to the kernel
        def trap_spawn(coro):
            task = new_task(coro)
            # task.parentid = current.id
            current._trap_result = task

        # ----------------------------------------
        # Cancel a task
        def trap_cancel_task(task, exc=TaskCancelled, val=None):
            if task.cancelled:
                return

            task.cancelled = True

            # Cancelling a task also cancels any currently pending timeout.
            # If a task is being cancelled, the delivery of a timeout is
            # somewhat immaterial--the task is already being cancelled.
            task.timeout = None

            # Set the cancellation exception
            if isinstance(exc, BaseException):
                task.cancel_pending = exc
            else:
                task.cancel_pending = exc(exc.__name__ if val is None else val)

            # If the task doesn't allow the delivery of a cancellation exception right now
            # we're done.  It's up to the task to check for it later
            if not task.allow_cancel:
                return

            # If the task doesn't have a cancellation function set, it means the task
            # is on the ready-queue.  It's not safe to deliver a cancellation exception
            # to it right now.  Instead, we simply return.  It will get cancelled
            # the next time it performs a blocking operation
            if not task.cancel_func:
                return

            # Cancel and reschedule the task
            task.cancel_func()
            task._trap_result = task.cancel_pending
            reschedule_task(task)
            task.cancel_pending = None

        # ----------------------------------------
        # Wait on a scheduler primitive
        def trap_sched_wait(sched, state):
            if check_cancellation():
                return
            suspend_task(state, sched._kernel_suspend(current))

        # ----------------------------------------
        # Reschedule one or more tasks from a scheduler primitive
        def trap_sched_wake(sched, n):
            tasks = sched._kernel_wake(n)
            for task in tasks:
                reschedule_task(task)

        # ----------------------------------------
        # Return the current value of the kernel clock
        def trap_clock():
            current._trap_result = time_monotonic()

        # ----------------------------------------
        # Sleep for a specified period. Returns value of monotonic clock.
        def trap_sleep(clock):
            nonlocal current
            if check_cancellation():
                return

            if clock <= 0:
                reschedule_task(current)
                current._trap_result = time_monotonic()
                current = None
                return

            set_timeout(clock + time_monotonic(), 'sleep')
            suspend_task('TIME_SLEEP',
                          lambda task=current: (sleepq.cancel((task.id, 'sleep'), task.sleep), setattr(task, 'sleep', None)))

        # ----------------------------------------
        # Set a timeout to be delivered to the calling task
        def trap_set_timeout(timeout):
            old_timeout = current.timeout
            if timeout is None:
                # If no timeout period is given, leave the current timeout in effect
                pass
            else:
                set_timeout(timeout)
                if old_timeout and current.timeout > old_timeout:
                    current.timeout = old_timeout
            current._trap_result = old_timeout

        # ----------------------------------------
        # Clear a previously set timeout
        def trap_unset_timeout(previous):
            # Here's an evil corner case.  Suppose the previous timeout in effect
            # has already expired?  If so, then we need to arrange for a timeout
            # to be generated.  However, this has to happen on the *next* blocking
            # call, not on this trap.  That's because the "unset" timeout feature
            # is usually done in the finalization stage of the previous timeout
            # handling.  If we were to raise a TaskTimeout here, it would get mixed
            # up with the prior timeout handling and all manner of head-explosion
            # will occur.

            set_timeout(None)
            current._trap_result = now = time_monotonic()
            if previous and previous >= 0 and previous < now:
                # Perhaps create a TaskTimeout pending exception here.
                set_timeout(previous)
            else:
                set_timeout(previous)
                current.timeout = previous
                # But there's one other evil corner case.  It's possible that
                # a timeout could be reset while a TaskTimeout exception
                # is pending.  If that happens, it means that the task has
                # left the timeout block.   We should probably take away the
                # pending exception.
                if isinstance(current.cancel_pending, TaskTimeout):
                    current.cancel_pending = None

        # ----------------------------------------
        # Return the running kernel
        def trap_get_kernel():
            current._trap_result = kernel

        # ----------------------------------------
        # Return the currently running task
        def trap_get_current():
            current._trap_result = current

        # ------------------------------------------------------------
        # Final setup.
        # ------------------------------------------------------------

        # Create the traps tables
        kernel._traps = traps = { key:value for key, value in locals().items()
                                  if key.startswith('trap_') }

        # Initialize activations
        kernel._activations = _activations = \
            [ act() if (isinstance(act, type) and issubclass(act, Activation)) else act
                    for act in kernel._activations ]

        for act in _activations:
            act.activate(kernel)

        # Initialize the loopback task (if not already initialized)
        init_loopback()
        task = new_task(_kernel_task())
        task.daemon = True

        # ------------------------------------------------------------
        # Main Kernel Loop.  Runs the supplied coroutine until it
        # terminates. If no coroutine is supplied, it runs one cycle
        # of the kernel.
        # ------------------------------------------------------------
        def kernel_run(coro):
            nonlocal current
            main_task = new_task(coro) if coro else None
            del coro
            trap = None

            while True:
                # ------------------------------------------------------------
                # I/O Polling/Waiting
                # ------------------------------------------------------------

                if ready or not main_task:
                    timeout = 0
                else:
                    current_time = time_monotonic()
                    timeout = sleepq.next_deadline(current_time)
                    if selector_max_timeout and (timeout is None or timeout > selector_max_timeout):
                        timeout = selector_max_timeout
                try:
                    events = selector_select(timeout)
                except OSError as e:
                    # If there is nothing to select, windows throws an
                    # OSError, so just set events to an empty list.
                    wsaeinval = getattr(errno, 'WSAEINVAL', None)
                    einval = getattr(errno, 'EINVAL', None)
                    if e.errno not in (wsaeinval, einval):
                        raise
                    events = []

                # Reschedule tasks with completed I/O
                for key, mask in events:
                    rtask, wtask = key.data
                    emask = key.events
                    intfd = isinstance(key.fileobj, int)
                    if mask & EVENT_READ:
                        # Discussion: If the associated fileobj is *not* a
                        # bare integer file descriptor, we keep a record
                        # of the last I/O event in _last_io and leave the
                        # task registered on the event loop.  If it
                        # performs the same I/O operation again, it will
                        # get a speed boost from not having to re-register
                        # its event. However, it's not safe to use this
                        # optimization with bare integer fds.  These fds
                        # often get reused and there is a possibility that
                        # a fd will get closed and reopened on a different
                        # resource without it being detected by the
                        # kernel.  For that case, its critical that we not
                        # leave the fd on the event loop.
                        rtask._last_io = None if intfd else (key.fileobj, EVENT_READ)
                        reschedule_task(rtask)
                        emask &= ~EVENT_READ
                        rtask = None

                    if mask & EVENT_WRITE:
                        wtask._last_io = None if intfd else (key.fileobj, EVENT_WRITE)
                        reschedule_task(wtask)
                        emask &= ~EVENT_WRITE
                        wtask = None

                    # Unregister the task if fileobj is not an integer fd (see
                    # note above).
                    if intfd:
                        if emask:
                            selector_modify(key.fileobj, emask, (rtask, wtask))
                        else:
                            selector_unregister(key.fileobj)


                # ------------------------------------------------------------
                # Time handling (sleep/timeouts)
                # ------------------------------------------------------------

                current_time = time_monotonic()
                for tm, (taskid, sleep_type) in sleepq.expired(current_time):
                    # When a task wakes, verify that the timeout value matches that stored
                    # on the task. If it differs, it means that the task completed its
                    # operation, was cancelled, or is no longer concerned with this
                    # sleep operation.  In that case, we do nothing
                    task = tasks.get(taskid)

                    if task is None:
                        continue
                    if tm != getattr(task, sleep_type):
                        continue

                    setattr(task, sleep_type, None)

                    if sleep_type == 'sleep':
                        task._trap_result = current_time
                        reschedule_task(task)

                    # If cancellation is allowed and the task is blocked, reschedule it
                    elif task.allow_cancel and task.cancel_func:
                        task.cancel_func()
                        task._trap_result = TaskTimeout(current_time)
                        reschedule_task(task)

                    # Task is on the ready queue or can't be cancelled right now;
                    # mark it as pending cancellation
                    else:
                        task.cancel_pending = TaskTimeout(current_time)

                # ------------------------------------------------------------
                # Run ready tasks
                # ------------------------------------------------------------

                for _ in range(len(ready)):
                    active = current = ready_popleft()
                    for a in _activations:
                        a.running(active)
                    active.state = 'RUNNING'
                    active.cycles += 1

                    # The current task runs until it suspends or terminates
                    while current:
                        try:
                            trap = current.send(current._trap_result)
                        except BaseException as e:
                            # If any exception has occurred, the task is done.
                            current = None

                            # Wake all joining tasks and enter the terminated state.
                            for wtask in active.joining._kernel_wake(len(active.joining)):
                                reschedule_task(wtask)
                            active.terminated = True
                            active.state = 'TERMINATED'
                            del tasks[active.id]
                            active.timeout = None
                            # Normal termination (set the result)
                            if isinstance(e, StopIteration):
                                active.result = e.value
                            else:
                                # Abnormal termination (set an exception)
                                active.exception = e
                                if (active != main_task and not isinstance(e, (CancelledError, SystemExit))):
                                    log.error('Task Crash: %r', active, exc_info=True)
                                if not isinstance(e, (Exception, CancelledError)):
                                    raise
                            break

                        # Run the trap function.  This is never supposed to raise
                        # an exception unless there's a fatal programming error in
                        # the kernel itself.  Such errors cause Curio to die. They
                        # are not reported back to tasks.
                        current._trap_result = None
                        try:
                            traps[trap[0]](*trap[1:])
                        except:
                            # Disable any further use of the kernel on fatal crash.
                            kernel._shutdown_funcs = None
                            raise

                    # --- The active task has suspended

                    # Unregister any prior I/O listening
                    if active._last_io:
                        unregister_event(*active._last_io)
                        active._last_io = None

                    # Trigger scheduler activations (if any)
                    for a in _activations:
                        a.suspended(active, trap)
                        if active.terminated:
                            a.terminated(active)
                    current = active = trap = None

                # If the main task has terminated, we're done.
                if main_task:
                    if main_task.terminated:
                        main_task.joined = True
                        return main_task
                else:
                    return None

        return kernel_run


def run(corofunc, *args, with_monitor=False, selector=None,
        debug=None, activations=None, **kernel_extra):
    '''
    Run the curio kernel with an initial task and execute until all
    tasks terminate.  Returns the task's final result (if any). This
    is a convenience function that should primarily be used for
    launching the top-level task of a curio-based application.  It
    creates an entirely new kernel, runs the given task to completion,
    and concludes by shutting down the kernel, releasing all resources used.

    Don't use this function if you're repeatedly launching a lot of
    new tasks to run in curio. Instead, create a Kernel instance and
    use its run() method instead.
    '''
    kernel = Kernel(selector=selector, debug=debug, activations=activations,
                    **kernel_extra)

    # Check if a monitor has been requested
    if with_monitor or 'CURIOMONITOR' in os.environ:
        from .monitor import Monitor
        m = Monitor(kernel)
        kernel._call_at_shutdown(m.close)
        kernel.run(m.start)

    with kernel:
        return kernel.run(corofunc, *args)

# An Activation is used to monitor and effect what happens
# during task execution in the Curio kernel. They are often used to
# implement tracers, debuggers, and other diagonistic tools.
# See curio/debug.py for some specific examples.

class Activation:

    def activate(self, kernel):
        '''
        Called each time the kernel sets up its environment and is ready to run.
        kernel is an instance of the kernel that's executing.
        '''

    def created(self, task):
        '''
        Called immediately after a task has been created.
        '''

    def running(self, task):
        '''
        Called right before the next execution cycle of a task.
        '''

    def suspended(self, task, trap):
        '''
        Called after the task has suspended due to a trap.
        '''

    def terminated(self, task):
        '''
        Called after a task has terminated, but prior to the task
        being collected by any associated join() operation.
        '''



# curio/subprocess.py
#
# A curio-compatible standin for the subprocess module.  Provides
# asynchronous compatible versions of Popen(), check_output(),
# and run() functions.

__all__ = ['run', 'Popen', 'CompletedProcess', 'CalledProcessError',
           'SubprocessError', 'check_output', 'PIPE', 'STDOUT', 'DEVNULL']

# -- Standard Library

import subprocess
import os
import sys

from subprocess import (
    CompletedProcess,
    SubprocessError,
    CalledProcessError,
    PIPE,
    STDOUT,
    DEVNULL,
)

# -- Curio

from .task import spawn
from .time import sleep
from .errors import CancelledError
from .io import FileStream
from . import thread
from .workers import run_in_thread

if sys.platform.startswith('win'):
    from .file import AsyncFile as FileStream

class Popen(object):
    '''
    Curio wrapper around the Popen class from the subprocess module. All of the
    methods from subprocess.Popen should be available, but the associated file
    objects for stdin, stdout, stderr have been replaced by async versions.
    Certain blocking operations (e.g., wait() and communicate()) have been
    replaced by async compatible implementations.   Explicit timeouts
    are not available. Use the timeout_after() function for timeouts.
    '''

    def __init__(self, args, **kwargs):
        if 'universal_newlines' in kwargs:
            raise RuntimeError('universal_newlines argument not supported')

        # If stdin has been given and it's set to a curio FileStream object,
        # then we need to flip it to blocking.
        if 'stdin' in kwargs:
            stdin = kwargs['stdin']
            if isinstance(stdin, FileStream):
                # At hell's heart I stab thy coroutine attempting to read from a stream
                # that's been used as a pipe input to a subprocess.  Must set back to
                # blocking or all hell breaks loose in the child.
                if hasattr(os, 'set_blocking'):
                    os.set_blocking(stdin.fileno(), True)

        self._popen = subprocess.Popen(args, **kwargs)

        if self._popen.stdin:
            self.stdin = FileStream(self._popen.stdin)
        if self._popen.stdout:
            self.stdout = FileStream(self._popen.stdout)
        if self._popen.stderr:
            self.stderr = FileStream(self._popen.stderr)

    def __getattr__(self, name):
        return getattr(self._popen, name)

    async def wait(self):
        retcode = self._popen.poll()
        if retcode is None:
            retcode = await run_in_thread(self._popen.wait)
        return retcode

    async def communicate(self, input=b''):
        '''
        Communicates with a subprocess.  input argument gives data to
        feed to the subprocess stdin stream.  Returns a tuple (stdout, stderr)
        corresponding to the process output.  If cancelled, the resulting
        cancellation exception has stdout_completed and stderr_completed
        attributes attached containing the bytes read so far.
        '''
        stdout_task = await spawn(self.stdout.readall) if self.stdout else None
        stderr_task = await spawn(self.stderr.readall) if self.stderr else None
        try:
            if input:
                await self.stdin.write(input)
                await self.stdin.close()

            stdout = await stdout_task.join() if stdout_task else b''
            stderr = await stderr_task.join() if stderr_task else b''
            return (stdout, stderr)
        except CancelledError as err:
            if stdout_task:
                await stdout_task.cancel()
                err.stdout = stdout_task.exception.bytes_read
            else:
                err.stdout = b''

            if stderr_task:
                await stderr_task.cancel()
                err.stderr = stderr_task.exception.bytes_read
            else:
                err.stderr = b''
            raise

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        if self.stdout:
            await self.stdout.close()

        if self.stderr:
            await self.stderr.close()

        if self.stdin:
            await self.stdin.close()

        # Wait for the process to terminate
        await self.wait()

    def __enter__(self):
        return thread.AWAIT(self.__aenter__())

    def __exit__(self, *args):
        return thread.AWAIT(self.__aexit__(*args))


async def run(args, *, stdin=None, input=None, stdout=None, stderr=None, shell=False, check=False):
    '''
    Curio-compatible version of subprocess.run()
    '''
    if input:
        stdin = subprocess.PIPE
    else:
        stdin = None

    async with Popen(args, stdin=stdin, stdout=stdout, stderr=stderr, shell=shell) as process:
        try:
            stdout, stderr = await process.communicate(input)
        except CancelledError as err:
            process.kill()
            stdout, stderr = await process.communicate()
            # Append the remaining stdout, stderr to the exception
            err.stdout += stdout
            err.stderr += stderr
            raise err
        except:
            process.kill()
            raise

    retcode = process.poll()
    if check and retcode:
        raise CalledProcessError(retcode, process.args,
                                 output=stdout, stderr=stderr)
    return CompletedProcess(process.args, retcode, stdout, stderr)


async def check_output(args, *, stdin=None, stderr=None, shell=False, input=None):
    '''
    Curio compatible version of subprocess.check_output()
    '''
    out = await run(args, stdout=PIPE, stdin=stdin, stderr=stderr, shell=shell,
                    check=True, input=input)
    return out.stdout



# curio/time.py
#
# Functionality related to time handling including timeouts and sleeping

__all__ = [
    'clock', 'sleep', 'timeout_after', 'ignore_after',
    ]

# -- Standard library

import logging
log = logging.getLogger(__name__)

# --- Curio

from .task import current_task
from .traps import *
from .errors import *
from . import meta

async def clock():
    '''
    Immediately return the current value of the kernel clock. There
    are no side-effects such as task preemption or cancellation.
    '''
    return await _clock()

async def sleep(seconds):
    '''
    Sleep for a specified number of seconds.  Sleeping for 0 seconds
    makes a task immediately switch to the next ready task (if any).
    Returns the value of the kernel clock when awakened.
    '''
    return await _sleep(seconds)

class _TimeoutAfter(object):
    '''
    Helper class used by timeout_after() and ignore_after() functions
    when used as a context manager.  For example:

        async with timeout_after(delay):
            statements
            ...
    '''

    def __init__(self, clock, ignore=False, timeout_result=None):
        self._clock = clock
        self._ignore = ignore
        self._timeout_result = timeout_result
        self.expired = False
        self.result = True

    async def __aenter__(self):
        task = await current_task()
        # Clock adjusted to absolute time
        if self._clock is not None:
            self._clock += await _clock()
        self._deadlines = task._deadlines
        self._deadlines.append(self._clock)
        self._prior = await _set_timeout(self._clock)
        return self

    async def __aexit__(self, ty, val, tb):
        current_clock = await _unset_timeout(self._prior)

        # Discussion.  If a timeout has occurred, it will either
        # present itself here as a TaskTimeout or TimeoutCancellationError
        # exception.  The value of this exception is set to the current
        # kernel clock which can be compared against our own deadline.
        # What happens next is driven by these rules:
        #
        # 1.  If we are the outer-most context where the timeout
        #     period has expired, then a TaskTimeout is raised.
        #
        # 2.  If the deadline has expired for at least one outer
        #     context, (but not us), a TimeoutCancellationError is
        #     raised.  This means that time has expired elsewhere.
        #     We're being cancelled because of that, but the reason
        #     for the cancellation wasn't due to a timeout on our
        #     part.
        #
        # 3.  If the timeout period has not expired on ANY remaining
        #     timeout context, it means that a timeout has escaped
        #     some inner timeout context where it should have been
        #     caught. This is an operational error.  We raise
        #     UncaughtTimeoutError.

        try:
            if ty in (TaskTimeout, TimeoutCancellationError):
                timeout_clock = val.args[0]
                # Find the outer most deadline that has expired
                for n, deadline in enumerate(self._deadlines):
                    if deadline <= timeout_clock:
                        break
                else:
                    # No remaining context has expired. An operational error
                    raise UncaughtTimeoutError('Uncaught timeout received')

                if n < len(self._deadlines) - 1:
                    if ty is TaskTimeout:
                        raise TimeoutCancellationError(val.args[0]).with_traceback(tb) from None
                    else:
                        return False
                else:
                    # The timeout is us.  Make sure it's a TaskTimeout (unless ignored)
                    self.result = self._timeout_result
                    self.expired = True
                    if self._ignore:
                        return True
                    else:
                        if ty is TimeoutCancellationError:
                            raise TaskTimeout(val.args[0]).with_traceback(tb) from None
                        else:
                            return False
            elif ty is None:
                if current_clock > self._deadlines[-1]:
                    # Further discussion.  In the presence of threads and blocking
                    # operations, it's possible that a timeout has expired, but
                    # there was simply no opportunity to catch it because there was
                    # no suspension point.
                    badness = current_clock - self._deadlines[-1]
                    log.warning('%r. Operation completed successfully, '
                                'but it took longer than an enclosing timeout. Badness delta=%r.',
                                await current_task(), badness)

        finally:
            self._deadlines.pop()

    def __enter__(self):
        return thread.AWAIT(self.__aenter__())

    def __exit__(self, *args):
        return thread.AWAIT(self.__aexit__(*args))

async def _timeout_after_func(clock, coro, args,
                              ignore=False, timeout_result=None):
    coro = meta.instantiate_coroutine(coro, *args)
    async with _TimeoutAfter(clock, ignore=ignore, timeout_result=timeout_result):
        return await coro

def timeout_after(seconds, coro=None, *args):
    '''
    Raise a TaskTimeout exception in the calling task after seconds
    have elapsed.  This function may be used in two ways. You can
    apply it to the execution of a single coroutine:

         await timeout_after(seconds, coro(args))

    or you can use it as an asynchronous context manager to apply
    a timeout to a block of statements:

         async with timeout_after(seconds):
             await coro1(args)
             await coro2(args)
             ...
    '''
    if coro is None:
        return _TimeoutAfter(seconds)
    else:
        return _timeout_after_func(seconds, coro, args)

def ignore_after(seconds, coro=None, *args, timeout_result=None):
    '''
    Stop the enclosed task or block of code after seconds have
    elapsed.  No exception is raised when time expires. Instead, None
    is returned.  This is often more convenient that catching an
    exception.  You can apply the function to a single coroutine:

        if await ignore_after(5, coro(args)) is None:
            # A timeout occurred
            ...

    Alternatively, you can use this function as an async context
    manager on a block of statements like this:

        async with ignore_after(5) as r:
            await coro1(args)
            await coro2(args)
            ...
        if r.result is None:
            # A timeout occurred

    When used as a context manager, the return manager object has
    a result attribute that will be set to None if the time
    period expires (or True otherwise).

    You can change the return result to a different value using
    the timeout_result keyword argument.
    '''
    if coro is None:
        return _TimeoutAfter(seconds, ignore=True, timeout_result=timeout_result)
    else:
        return _timeout_after_func(seconds, coro, args, ignore=True, timeout_result=timeout_result)

from . import thread



# curio/meta.py
#     ___
#     \./      DANGER:  This module implements some experimental
#  .--.O.--.            metaprogramming techniques involving async/await.
#   \/   \/             If you use it, you might die. No seriously.
#

__all__ = [
    'iscoroutinefunction', 'finalize', 'awaitable', 'asyncioable',
    'curio_running', 'instantiate_coroutine', 'from_coroutine',
 ]

# -- Standard Library

from sys import _getframe
import sys
import inspect
from functools import wraps, partial
import dis
import asyncio
import threading
from contextlib import contextmanager
import collections.abc

# -- Curio

from .errors import SyncIOError


_locals = threading.local()

# Context manager that is used when the kernel is executing.

@contextmanager
def running(kernel):
    if getattr(_locals, 'running', False):
        raise RuntimeError('Only one Curio kernel per thread is allowed')
    _locals.running = True
    _locals.kernel = kernel
    try:
        with asyncgen_manager():
            yield
    finally:
        _locals.running = False
        _locals.kernel = None

def curio_running():
    '''
    Return a flag that indicates whether or not Curio is running in the current thread.
    '''
    return getattr(_locals, 'running', False)

_CO_NESTED = inspect.CO_NESTED
_CO_FROM_COROUTINE = inspect.CO_COROUTINE | inspect.CO_ITERABLE_COROUTINE | inspect.CO_ASYNC_GENERATOR
_isasyncgenfunction = inspect.isasyncgenfunction

def from_coroutine(level=2, _cache={}):
    f_code = _getframe(level).f_code
    if f_code in _cache:
        return _cache[f_code]
    if f_code.co_flags & _CO_FROM_COROUTINE:
        _cache[f_code] = True
        return True
    else:
        # Comment:  It's possible that we could end up here if one calls a function
        # from the context of a list comprehension or a generator expression. For
        # example:
        #
        #   async def coro():
        #        ...
        #        a = [ func() for x in s ]
        #        ...
        #
        # Where func() is some function that we've wrapped with one of the decorators
        # below.  If so, the code object is nested and has a name such as <listcomp> or <genexpr>
        if (f_code.co_flags & _CO_NESTED and f_code.co_name[0] == '<'):
            return from_coroutine(level + 2)
        else:
            _cache[f_code] = False
            return False

def iscoroutinefunction(func):
    '''
    Modified test for a coroutine function with awareness of functools.partial
    '''
    if isinstance(func, partial):
        return iscoroutinefunction(func.func)
    if hasattr(func, '__func__'):
        return iscoroutinefunction(func.__func__)
    return inspect.iscoroutinefunction(func) or hasattr(func, '_awaitable') or _isasyncgenfunction(func)

def instantiate_coroutine(corofunc, *args, **kwargs):
    '''
    Try to instantiate a coroutine. If corofunc is already a coroutine,
    we're done.  If it's a coroutine function, we call it inside an
    async context with the given arguments to create a coroutine.  If
    it's not a coroutine, we call corofunc(*args, **kwargs) and hope
    for the best.
    '''
    if isinstance(corofunc, collections.abc.Coroutine) or inspect.isgenerator(corofunc):
        assert not args and not kwargs, "arguments can't be passed to an already instantiated coroutine"
        return corofunc

    if not iscoroutinefunction(corofunc) and not getattr(corofunc, '_async_thread', False):
        coro = corofunc(*args, **kwargs)
        if not isinstance(coro, collections.abc.Coroutine):
            raise TypeError(f'Could not create coroutine from {corofunc}')
        return coro

    async def context():
        return corofunc(*args, **kwargs)

    try:
        context().send(None)
    except StopIteration as e:
        return e.value

def awaitable(syncfunc):
    '''
    Decorator that allows an asynchronous function to be paired with a
    synchronous function in a single function call.  The selection of
    which function executes depends on the calling context.  For example:

        def spam(sock, maxbytes):                       (A)
            return sock.recv(maxbytes)

        @awaitable(spam)                                (B)
        async def spam(sock, maxbytes):
            return await sock.recv(maxbytes)

    In later code, you could use the spam() function in either a synchronous
    or asynchronous context.  For example:

        def foo():
            ...
            r = spam(s, 1024)          # Calls synchronous function (A) above
            ...

        async def bar():
            ...
            r = await spam(s, 1024)    # Calls async function (B) above
            ...

    '''
    def decorate(asyncfunc):
        if inspect.signature(syncfunc) != inspect.signature(asyncfunc):
            raise TypeError(f'{syncfunc.__name__} and async {asyncfunc.__name__} have different signatures')

        @wraps(asyncfunc)
        def wrapper(*args, **kwargs):
            if from_coroutine():
                return asyncfunc(*args, **kwargs)
            else:
                return syncfunc(*args, **kwargs)
        wrapper._syncfunc = syncfunc
        wrapper._asyncfunc = asyncfunc
        wrapper._awaitable = True
        wrapper.__doc__ = syncfunc.__doc__ or asyncfunc.__doc__
        return wrapper
    return decorate

def asyncioable(awaitablefunc):
    '''
    Decorator that additionally allows an asyncio compatible call to
    be attached to an already awaitable function. For example:

      def spam():
          print('Synchronous spam')

      @awaitable(spam)
      def spam():
          print('Async spam (Curio)')

      @asynioable(spam)
      def spam():
          print('Async spam (asyncio)')

    This only works if Curio/Asyncio are running in different threads.
    Main use is in the implementation of UniversalQueue.
    '''
    def decorate(asyncfunc):
        @wraps(asyncfunc)
        def wrapper(*args, **kwargs):
            if from_coroutine():
                # Check if we're Curio or not
                if curio_running():
                    return awaitablefunc._asyncfunc(*args, **kwargs)
                else:
                    return asyncfunc(*args, **kwargs)
            else:
                return awaitablefunc._syncfunc(*args, **kwargs)
        wrapper._awaitable = True
        return wrapper
    return decorate

class finalize(object):
    '''
    Context manager that safely finalizes an asynchronous generator.
    This might be needed if an asynchronous generator uses async functions
    in try-finally and other constructs.
    '''
    def __init__(self, aobj):
        self.aobj = aobj

    async def __aenter__(self):
        return self.aobj

    async def __aexit__(self, ty, val, tb):
        if hasattr(self.aobj, 'aclose'):
            await self.aobj.aclose()


# This context manager is used to manage the execution of async generators
# in Python 3.6.  In certain circumstances, they can't be used safely
# unless finalized properly.  This context manager installs some a hook
# for detecting lack of finalization.

@contextmanager
def asyncgen_manager():
    if hasattr(sys, 'get_asyncgen_hooks'):
        old_asyncgen_hooks = sys.get_asyncgen_hooks()
        def _fini_async_gen(agen):
            if agen.ag_frame is not None:
                raise RuntimeError("Async generator with async finalization must be wrapped by\n"
                                   "async with curio.meta.finalize(agen) as agen:\n"
                                   "    async for n in agen:\n"
                                   "         ...\n"
                                   "See PEP 533 for further discussion.")

        sys.set_asyncgen_hooks(None, _fini_async_gen)
    try:
        yield
    finally:
        if hasattr(sys, 'get_asyncgen_hooks'):
            sys.set_asyncgen_hooks(*old_asyncgen_hooks)




# curio/file.py
#
# Let's talk about files for a moment.  Suppose you're in a coroutine
# and you start using things like the built-in open() function:
#
#     async def coro():
#         f = open(somefile, 'r')
#         data = f.read()
#         ...
#
# Yes, it will "work", but who knows what's actually going to happen
# on that open() call and associated read().  If it's on disk, the
# whole program might lock up for a few milliseconds (aka. "an
# eternity") doing a disk seek.  While that happens, your whole
# coroutine based server is going to grind to a screeching halt.  This
# is bad--especially if a lot of coroutines start doing it all at
# once.
#
# Knowing how to handle this is a tricky question.  Traditional files
# don't really support "async" in the usual way a socket might. You
# might be able to do something sneaky with asynchronous POSIX APIs
# (i.e., aio_* functions) or maybe thread pools.  However, one thing
# is for certain--if files are going to be handled in a sane way, they're
# going to have an async interface.
#
# This file does just that by providing an async-compatible aopen()
# call.  You use it the same way you use open() and a normal file:
#
#    async def coro():
#        async with aopen(somefile, 'r') as f:
#            data = await f.read()
#            ...
#
# If you want to use iteration, make sure you use the asynchronous version:
#
#    async def coro():
#        async with aopen(somefile, 'r') as f:
#            async for line in f:
#                ...
#

__all__ = ['aopen', 'anext']

# -- Standard library

from contextlib import contextmanager
from functools import partial

# -- Curio

from .workers import run_in_thread
from .errors import SyncIOError, CancelledError
from . import thread

class AsyncFile(object):
    '''
    An async wrapper around a standard file object.  Uses threads to
    execute various I/O operations in a way that avoids blocking
    the Curio kernel loop.
    '''

    def __init__(self, fileobj, open_args=None, open_kwargs=None):
        self._fileobj = fileobj
        self._open_args = open_args
        self._open_kwargs = open_kwargs

    def __repr__(self):
        return 'AsyncFile(%r)' % self._fileobj

    @contextmanager
    def blocking(self):
        '''
        Expose the underlying file in blocking mode for use with synchronous code.
        '''
        yield self._file

    @property
    def _file(self):
        if self._fileobj is None:
            raise RuntimeError('Must use an async file as an async-context-manager.')
        return self._fileobj

    async def read(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.read, *args, **kwargs))

    async def read1(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.read1, *args, **kwargs))

    async def readinto(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.readinto, *args, **kwargs))

    async def readinto1(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.readinto1, *args, **kwargs))

    async def readline(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.readline, *args, **kwargs))

    async def readlines(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.readlines, *args, **kwargs))

    async def write(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.write, *args, **kwargs))

    async def writelines(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.writelines, *args, **kwargs))

    async def flush(self):
        return await run_in_thread(self._file.flush)

    async def close(self):
        return await run_in_thread(self._file.close)

    async def seek(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.seek, *args, **kwargs))

    async def tell(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.tell, *args, **kwargs))

    async def truncate(self, *args, **kwargs):
        return await run_in_thread(partial(self._file.truncate, *args, **kwargs))

    def __iter__(self):
        raise SyncIOError('Use asynchronous iteration')

    def __next__(self):
        raise SyncIOError('Use asynchronous iteration')

    def __enter__(self):
        return thread.AWAIT(self.__aenter__())

    def __exit__(self, *args):
        return thread.AWAIT(self.__aexit__(*args))

    def __aiter__(self):
        return self

    async def __aenter__(self):
        if self._fileobj is None:
            self._fileobj = await run_in_thread(partial(open, *self._open_args, **self._open_kwargs))
        return self

    async def __aexit__(self, *args):
        await self.close()

    async def __anext__(self):
        data = await run_in_thread(next, self._file, None)
        if data is None:
            raise StopAsyncIteration
        return data

    def __getattr__(self, name):
        return getattr(self._file, name)

    # Compatibility with io.FileStream
    async def readall(self):
        chunks = []
        maxread = 65536
        sep = '' if hasattr(self._file, 'encoding') else b''
        while True:
            try:
                chunk = await self.read(maxread)
            except CancelledError as e:
                e.bytes_read = sep.join(chunks)
                raise
            if not chunk:
                return sep.join(chunks)
            chunks.append(chunk)
            if len(chunk) == maxread:
                maxread *= 2

def aopen(*args, **kwargs):
    '''
    Async version of the builtin open() function that returns an async-compatible
    file object.  Takes the same arguments.  Returns a wrapped file in which
    blocking I/O operations must be awaited.
    '''
    return AsyncFile(None, args, kwargs)

async def anext(f, sentinel=object):
    '''
    Async version of the builtin next() function that advances an async iterator.
    Sometimes used to skip a single line in files.
    '''
    try:
        return await f.__anext__()
    except StopAsyncIteration:
        if sentinel is not object:
            return sentinel
        else:
            raise



# curio/ssl.py
#
# Wrapper around built-in SSL module

__all__ = []

# -- Standard Library

from functools import wraps, partial

try:
    import ssl as _ssl
    from ssl import *
except ImportError:
    _ssl = None

    # We need these exceptions defined, even if ssl is not available.
    class SSLWantReadError(Exception):
        pass

    class SSLWantWriteError(Exception):
        pass

# -- Curio

from .workers import run_in_thread
from .io import Socket

if _ssl:
    @wraps(_ssl.wrap_socket)
    async def wrap_socket(sock, *args, do_handshake_on_connect=True, **kwargs):
        if isinstance(sock, Socket):
            sock = sock._socket

        ssl_sock = _ssl.wrap_socket(sock, *args, do_handshake_on_connect=False, **kwargs)
        cssl_sock = Socket(ssl_sock)
        cssl_sock.do_handshake_on_connect = do_handshake_on_connect
        if do_handshake_on_connect and ssl_sock._connected:
            await cssl_sock.do_handshake()
        return cssl_sock

    @wraps(_ssl.get_server_certificate)
    async def get_server_certificate(*args, **kwargs):
        return await run_in_thread(partial(_ssl.get_server_certificate, *args, **kwargs))

    # Small wrapper class to make sure the wrap_socket() method returns the right type
    class CurioSSLContext(object):

        def __init__(self, context):
            self._context = context

        def __getattr__(self, name):
            return getattr(self._context, name)

        async def wrap_socket(self, sock, *args, do_handshake_on_connect=True, **kwargs):
            sock = self._context.wrap_socket(
                sock._socket, *args, do_handshake_on_connect=False, **kwargs)
            csock = Socket(sock)
            csock.do_handshake_on_connect = do_handshake_on_connect
            if do_handshake_on_connect and sock._connected:
                await csock.do_handshake()
            return csock

        def __setattr__(self, name, value):
            if name == '_context':
                super().__setattr__(name, value)
            else:
                setattr(self._context, name, value)

    # Name alias
    def SSLContext(protocol):
        return CurioSSLContext(_ssl.SSLContext(protocol))

    @wraps(_ssl.create_default_context)
    def create_default_context(*args, **kwargs):
        context = _ssl.create_default_context(*args, **kwargs)
        return CurioSSLContext(context)
